/***************************************************************************
                          graph.cpp  -  description
                             -------------------
    begin                : Mon Mar 11 2002
    copyright            : (C) 2002 by Volker Nannen
    email                : volker@nannen.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "graph.h"
#include "project.h"
#include "collection.h"

#include <qcolor.h>
#include <qdom.h>

#include <cmath>

using namespace std;

Graph::Graph(Project * project, const QString name)
  : BasicElement(0, name),
    project(project),
    canCopy(false),
    toggled(false),
    points(0, project->getCanvas()),
    pointCount(0)
{}
Graph::Graph(Graph * other)
  : BasicElement(0, other->project->makeName(other->getElementName())),
    project(other->project),
    canCopy(false),
    toggled(false),
    points(0, project->getCanvas()),
    pointCount(0)
{
  xml_copy(other);
}
Graph::~Graph(){
  Graph::destroy();
}
void Graph::destroy(){
  points.clear();
}
bool Graph::xml_read(QDomElement * element){
  destroy();
  BasicElement::xml_read(element);
  if(element){
    QDomElement tmpElement = element->elementsByTagName("Graph").item(0).toElement(); // the points are called 'Graph'
    points.xml_read( &tmpElement );
    points.recalculate();
  }
  controlCopyBool = copyBool;
  setID( getID() );
  return true;
}
void Graph::setValid(const bool yes){
  valid = yes;
  if(points.getVisible() && ! valid){
    recalculate();
  }
}
void Graph::setID(const int newID){
  BasicElement::setID(newID);
  points.setID(newID);
}
void Graph::xml_list(){  // not called by Split::xml_list() !!!
  BasicElement::xml_list();
  points.setID(getID());
  xml_bool("copy", copyBool, false, (canCopy ? EDIT : HIDE));
  xml_bool("delete", deleteBool, false, EDIT);
  xml_int("plotID", plotID, -1, HIDE);
  xml_object( & points );
//  cerr << " name " << getID() << " points " << points.size() << endl;
}
void Graph::statistics(){
  sum = 0;
  sumOfSquares = 0;
  pointCount = points.size();
  for(Points::iterator iter = points.begin(); iter != points.end(); iter++){
    sum += (*iter)->getValueY();
    sumOfSquares += (*iter)->getValueY() * (*iter)->getValueY();
  }
  mean = sum / pointCount;
  sigma = sqrt(sumOfSquares / pointCount - mean * mean);
}
void Graph::refresh(){
  points.refresh();
  if( elementNameChanged() ){
    project->nameChanged(getID());
    Collection * parentCollection = getParentCollection();
    if( parentCollection ) {
      ScientificCanvas * plot = dynamic_cast<ScientificCanvas *>
                             ( parentCollection->IDToElement(plotID) );
      if( plot ) plot->setCaption(  project->getElementName()
                                  + "::" + getElementName() );
    }
  }
  if(controlCopyBool != copyBool){
    controlCopyBool = copyBool;
    project->addGraph(copyGraph());
  }
  if( deleteBool ){
    if( xml_warning( "Delete " + getElementName() + " ?" , true) ){
      deleteBool = false;
      emit valueChanged();
      return;
    }
    destroy();
    if(plotID > 0) project->deleteGraph(plotID);
    project->deleteGraph(getID());
    return;
  }
  setValid(valid);
}
void Graph::updateCanvas(){
  project->updateCanvas();
}
void Graph::toggleColor(const bool yes){
  if(yes && points.size() < 3000){
    int h, s, v;
    QColor toggle(points.getColor());
    toggle.hsv(&h, &s, &v);
    toggle.setHsv((h + 180) % 360, 255, v);
    setPen( QPen( toggle ) );
    toggled = true;
  } else if(toggled) {
    setPen( QPen( QColor(points.getColor()).rgb() ) );
    toggled = false;
  }
}
void Graph::setVisible(const bool yes){
  points.setVisible(yes);
  setValid(valid);
}
ScientificCanvas * Graph::makePlot(Scale::Method scaleMethod,
                                          double lowerLimit, double upperLimit,
                                          double leftLimit,  double rightLimit){
  Collection * parentCollection = getParentCollection();
  if( ! parentCollection ) return 0;
  ScientificCanvas * plot = dynamic_cast<ScientificCanvas *>
                         ( parentCollection->IDToElement(plotID) );
  if( ! plot ) {
    plot = new ScientificCanvas(project, getElementName() + " screen");
    project->addGraph(plot);
    plotID = plot->getID();
    plot->setSize(300, 300);
  } else {
    plot->clear();
  }
  plot->setCaption(project->getElementName() + "::" + getElementName() );
  Axes * axes = plot->getAxes();
  if ( ! axes ) return 0;
  axes->setLeftX(leftLimit);
  axes->setRightX(rightLimit);
  axes->setLowerY(lowerLimit);
  axes->setUpperY(upperLimit);
  axes->setVerticalScale( scaleMethod );
  plot->refresh();
  return plot;
}
