/***************************************************************************
                          learnpolynomial.cpp  -  description
                             -------------------
    begin                : ?' ??? 5 2002
    copyright            : (C) 2002 by Volker Nannen
    email                : volker@nannen.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "learnpolynomial.h"
#include "project.h"
#include "sample.h"
#include "scientificcanvas.h"
#include "axes.h"
#include "scale.h"
#include "collection.h"
#include "enhancedprogressbar.h"
#include "vandermonde.h"

#include <qmessagebox.h>

#include <algorithm>
//#include <algo.h>
#include <cmath>

using namespace std;

LearnPolynomial::LearnPolynomial(Project * project, const QString name, QDomElement * element)
  : Polynomial(project, name, element)
{}

LearnPolynomial::LearnPolynomial(LearnPolynomial * other)
  : Polynomial(other)
{}
void LearnPolynomial::refresh(){
  if(fit != controlFit && fit) fit = learn();
  Polynomial::refresh();
}
bool LearnPolynomial::learn(){
  Sample * sample = getSample(learnSampleID);
  if( ! sample || ! sample->getPoints()->size() ) {
    setOrder(0);
    if( ! sample ) xml_warning("no learn sample selected");
    else if( ! sample->getPoints()->size() ) xml_warning("empty learn sample");
    return true;
  }
  switch(method){
    case COMPLEX:
      return complexAnalysis( sample );
    case CROSS:
      return crossValidate( sample );
//    case TRIMMED_MDL:
//      return trimmed_MDL( sample );
    case ANALYSIS:
      return analysis( sample );
//    case ORDER_PRECISION:
//      orderPrecision( sample );
//      return false;
    case TWO_PART_MDL:
      return two_part_MDL( sample );
    case MIXTURE_MDL:
      return mixture_MDL( sample );
    case MANUAL:
      if( ! sample->getVanDerMonde()->leastSquare( order ) )
        return noMatrix();
      setOrder(order, precision);
//      makeParameter();
      return true;
  }
//  cerr << " learned \n";
  return true;
}
bool LearnPolynomial::complexAnalysis(Sample * sample){
            // for each learnSample size find the best fit.
  Points * allLearnPoints = sample->getPoints();
  int learnCount = allLearnPoints->size();
  Sample * testSample = getSample(testSampleID);
  if( ! testSample ) return noTest();
  Points * testPoints = sample->getPoints();
  if( testPoints->empty() ) return emptyTest();
  VanDerMonde * testVanDerMonde = testSample->getVanDerMonde();
  Points learnPoints;
  VanDerMonde learnVanDerMonde(&learnPoints);
  ScientificCanvas * plot = makePlot(Scale::SIMPLE, 0, maxOrder, 0, learnCount);
  if( ! plot ) return noPlot();
  Points * analysisPoints = plot->getNewPoints( getID(), "red" );
  double learnSizeStep = learnCount / sampleNumber;
  if( learnSizeStep < 1 ) learnSizeStep = 1;
  double degreeStep = maxOrder / stepNumber;
  if( degreeStep < 1 ) degreeStep = 1;
  int round = 0;
  int optDegree = 0;
  int maxDegree = learnCount < maxOrder ? learnCount : maxOrder;
  double sigma[maxOrder + 1];
  bool stop = false;
  EnhancedProgressBar progress( "complex analysis", learnCount+(int)learnSizeStep, 0);
  testVanDerMonde->setOrder(maxOrder);
  if( ! testVanDerMonde->processData( & progress ) )
    return noTestProcess();
  int index[learnCount];
  for(int i=0; i<learnCount; i++) index[i] = i;
  random_shuffle(index, index + (learnCount-1) );

  for(double mySize = learnSizeStep; mySize <= learnCount; mySize += learnSizeStep){
    int learnSize = int(mySize);
    optDegree = 0;
    int pointIndex   = 0;
    for(double degree = 0; degree < maxDegree; degree += degreeStep){
      sigma[int(degree)] = 0;
    }
    for(round=0; (round < 4) && ( round < learnCount / learnSize ); round++){ // use more than one
      learnPoints.destroy();                                                  // sample of size n
      learnVanDerMonde.setOrder(-1);
      for(int i=0; i < learnSize; i++, pointIndex++){      // build training set
        learnPoints.copy((*allLearnPoints)[index[pointIndex]]);
      }
      if( ! learnVanDerMonde.leastSquare( maxDegree, & progress ) )
        return noMatrix();
      for(double degree = 0; degree < maxDegree; degree += degreeStep){
        cerr << " learnSize " << learnPoints.size() << " \t degree " << degree << " \t round " << round <<endl;
        stop = progress.check(learnSize + int(learnSizeStep*learnSize/pointIndex));
        if( stop ) break;
        for(int i=0; i<= int(degree); i++) precisionArray[i] = int(degree/2);
        learnVanDerMonde.makeParameter(parameter, precisionArray, (int)degree);
        sigma[int(degree)] += testVanDerMonde->variance(parameter, (int)degree );
      }
      if(stop) break;
    }
    if(stop) break;
    double minSigma = -1;
    for(double degree = 0; degree < maxDegree; degree += degreeStep){
      sigma[int(degree)] /= round;
      if (sigma[int(degree)] < minSigma || minSigma < 0){
        optDegree = (int)degree;
        minSigma = sigma[(int)degree];
      }
    }
    Points * degreePoints = plot->getNewPoints( getID(), "green" );
    for(double degree = 0; degree < maxDegree; degree += degreeStep){
      double myVariance =   learnSize + learnSizeStep
                          * ( minSigma / sigma[int(degree)] - 1);
//      cerr << " === myVariance " << myVariance << " \t variances[" << int(degree) << "] = " << sigma[int(degree)] ;
      degreePoints->paint_back(myVariance, floor(degree));
    }
    analysisPoints->paint_back(learnSize, double(optDegree));
    plot->refresh();
//    cerr << " ----------------- learn size \t " << learnSize << " \t opt degree \t " << optDegree << endl;
  }
  setOrder(optDegree);
  return false;
}
bool LearnPolynomial::analysis(Sample * sample){
  VanDerMonde * learnVanDerMonde = sample->getVanDerMonde();
  int learnCount = learnVanDerMonde->getSampleSize();
  Sample * testSample = getSample(testSampleID);;
  if( ! testSample ) return noTest();
  VanDerMonde * testVanDerMonde = testSample->getVanDerMonde();
  Points * testPoints = testSample->getPoints();
  if( testPoints->empty() ) return emptyTest();
  setOrder(maxOrder);
  ScientificCanvas * plot = makePlot(Scale::LOGLOG, 0.00001, 100, 0, maxOrder);
  if( ! plot ) return noPlot();
  Points * analysisPoints = plot->getNewPoints( getID(), "red" );
  double minSigma(-1), maxSigma(0);
  int maxDegree = learnCount < maxOrder ? learnCount : maxOrder;
  int optDegree(0);
  double degreeStep = maxDegree / stepNumber;
  if( degreeStep < 1 ) degreeStep = 1;
  EnhancedProgressBar progress( "analysing best fit", maxDegree, 0);
  if( ! learnVanDerMonde->leastSquare( maxOrder, & progress ) )
    return noMatrix();
  testVanDerMonde->setOrder(maxOrder);
  if( ! testVanDerMonde->processData( & progress ) )
    return noTestProcess();
  cerr << " learnPolynomial: entering loop\n";
  cerr << " maxDegree = " << maxDegree << ", maxOrder = " << maxOrder
       << ", degreeStep = " << degreeStep << endl;
  for(double degree = 0; degree <= maxDegree; degree += degreeStep){
    if( progress.check(int(degree) ) ) break;
    for(int i=0; i<= int(degree); i++) precisionArray[i] = int(degree/2);
    learnVanDerMonde->makeParameter(parameter, precisionArray, (int)degree);
    double sigma = testVanDerMonde->variance( parameter, (int)degree );
    if (sigma < minSigma || minSigma < 0) {
      minSigma = sigma;
      optDegree = int(degree);
    }
    if (sigma > maxSigma) {
      maxSigma = sigma;
      plot->getAxes()->setUpperY(maxSigma);
    }
    analysisPoints->paint_back(double(degree), double(sigma));
    plot->refresh();
  }
  plot->getAxes()->setLowerY(minSigma * 0.9);
  plot->refresh();
  setOrder(optDegree);
  return true;
}
bool LearnPolynomial::mixture_MDL(Sample * sample){
  cerr << " mixture_MDL, m = 1.5 * d\n";
  Points * learnPoints = sample->getPoints();
  VanDerMonde * learnVanDerMonde = sample->getVanDerMonde();
  int count = learnPoints->size();
  if( count < (2*maxOrder) ) return xml_warning("for mixture MDL the number of points\n"
                                            "in the learn sample has to be at\n"
                                            "least 2 times that of the max order");
  double top(100);
  ScientificCanvas * plot = makePlot(Scale::SIMPLE, 1, top, 0, maxOrder);
  if( ! plot ) return noPlot();
  Points * analysisPoints = plot->getNewPoints( getID(), "red" );
  EnhancedProgressBar progress( "mixture MDL", maxOrder + 1, 0);
  if( ! learnVanDerMonde->leastSquare( maxOrder, & progress ) )
    return noMatrix();
  Points mPoints;
  {
    int index[count];
    for(int i=0; i<count; i++) index[i] = i;
    random_shuffle(index, index + (count-1) );
    for(int i=0; i < 1.5*maxOrder; i++) mPoints.copy(( * learnPoints)[index[i]]);
  }
  VanDerMonde mVanDerMonde( & mPoints );
  if( ! mVanDerMonde.leastSquare( maxOrder, & progress ) )
    return xml_warning("cant make matrix for auxiliary set");
  int m = mPoints.size();
  int n = count - m;
  int minLength(-1);
  int optDegree(0);
  double degreeStep = maxOrder / stepNumber;
  if( degreeStep < 1 ) degreeStep = 1;
  for(double myDegree=0; myDegree<=maxOrder; myDegree += degreeStep){
    int degree = int(myDegree);
    if ( progress.check(degree) ) break;
    double a1 = (m + n - degree) / 2.0;
    double a2 = (m - degree) / 2.0;
    for(int i=0; i<=degree; i++) precisionArray[i] = degree/2;



    learnVanDerMonde->makeParameter(parameter, precisionArray, degree);
    double log_a1 = a1 * log(n * learnVanDerMonde->variance( parameter, degree ) );

    mVanDerMonde.makeParameter(parameter, precisionArray, degree);
    double log_a2 = a2 * log(n * mVanDerMonde.variance( parameter, degree ) );

    bool detN, detM;
    double logDetN = learnVanDerMonde->logDeterminant(degree, detN);
    double logDetM = mVanDerMonde.logDeterminant(degree, detM);

    double log_det = 0.5 * ( logDetN - logDetM );

    double log_gamma = gamma(m, n, degree, a1, a2);

    double length = int(log_a1 - log_a2 + log_det - log_gamma);

    if( degree > 100) cerr << " deg " << degree << " \t det_N " << logDetN
         << " \t det_M " << logDetM
         << " \t log_a1 "  << log_a1 << " \t log_a2 "    << log_a2
         << " \t log_det " << log_det << " \t log_gamma " << log_gamma << endl;

    if( ! detM ) cerr << " det M is negative!\n";
    if( ! detN ) cerr << " det N is negative!\n";

//    if( (logN && !logM) || (!logN && logM) ){
//      length = 1e+300;
//    } else {
      if ( length < 1 ) length = 1;
      if( (length < minLength) || (minLength < 0) ){
        minLength = (int)length;
        optDegree = degree;
      }
      if (length > top) {
        top = length;
        plot->getAxes()->setUpperY(top);
      }
//    }
    analysisPoints->paint_back((double)degree, length);
    plot->refresh();
  }
  plot->getAxes()->setLowerY( minLength * 0.9 );
  plot->refresh();
  setOrder(optDegree);
  return true;
}
double LearnPolynomial::gamma(const int m, const int n, const int d,
                         const double a1, const double a2){
  return (   log(a1) * ( m+n-d-1 )
           - log(a2) * ( m-d-1 )
           - n ) / 2.0
         + 1.0 / ( 12.0 * a1 )
         - 1.0 / ( 12.0 * a2 )
         - 1.0 / ( 360.0 * pow(a1 ,3) )
         + 1.0 / ( 360.0 * pow(a2 ,3) )
         + 1.0 / ( 1260.0 * pow(a1 ,5) )
         - 1.0 / ( 1260.0 * pow(a2 ,5) );
}
bool LearnPolynomial::trimmed_MDL(Sample * sample){
  VanDerMonde * vanDerMonde = sample->getVanDerMonde();
  EnhancedProgressBar progress( "trimmed MDL", maxOrder + 1, 0);
  if( ! vanDerMonde->leastSquare( order, & progress ) )
    return noMatrix();
  setOrder(order, precision);
  vanDerMonde->makeParameter( parameter, precisionArray, order );
  descriptionLength = goodnessOfFit(variance, sample);
  int minLength = descriptionLength;

  for(int orderIndex = order; orderIndex >=0; orderIndex--){
    cerr << " --------------- order: " << orderIndex << endl;
    int minPrecision=0, maxPrecision;
    while(minPrecision != precisionArray[orderIndex]){
      if ( progress.check(order - orderIndex) ) return false;
      maxPrecision = precisionArray[orderIndex];
      precisionArray[orderIndex] = (minPrecision + precisionArray[orderIndex]) / 2;
      vanDerMonde->makeParameter(parameter, precisionArray, order);
      descriptionLength = goodnessOfFit(variance, sample);
      cerr << " --- precision: " << precisionArray[orderIndex]
           << ", \t--- length: " << descriptionLength  << endl;
      if(descriptionLength < minLength){
        minLength = descriptionLength;
      } else {
        minPrecision = precisionArray[orderIndex] + 1;
        precisionArray[orderIndex] = maxPrecision;
      }
    }
  }
//for(bool shrinking = true; shrinking; ;){
  bool shrinking=true;
  while(shrinking){
    cerr << " ===================== new round !!! ============\n";
    shrinking=false;
    for(int orderIndex = order; orderIndex >=0; orderIndex--){
      cerr << " --------------- order: " << orderIndex << endl;
      int bestPrecision = precisionArray[orderIndex];
      while(precisionArray[orderIndex]){
        if ( progress.check() ) return false;
        precisionArray[orderIndex]--;
        vanDerMonde->makeParameter(parameter, precisionArray, order);
        descriptionLength = goodnessOfFit(variance, sample);
        cerr << " --- precision: " << precisionArray[orderIndex] << ", \t--- length: " << descriptionLength  << endl;
        if(descriptionLength < minLength){
          minLength = descriptionLength;
          bestPrecision = precisionArray[orderIndex];
          shrinking = true;
        } else {
          if( (precisionArray[orderIndex] < 5) && (precisionArray[orderIndex] != 0) ){
            precisionArray[orderIndex] = 1;
          } else {
            precisionArray[orderIndex] = bestPrecision;
            break;
          }
        }
      }
    }
  }
  vanDerMonde->makeParameter(parameter, precisionArray, order);
  descriptionLength = goodnessOfFit(variance, sample);
  for(int i=0; i<= order; i++) cerr << " precision["<< i << "] = " << precisionArray[i] << endl;
  return true;
}
bool LearnPolynomial::two_part_MDL(Sample * sample){
  int bottom(0), top(1);
  ScientificCanvas * plot = makePlot(Scale::SIMPLE, bottom, top, 0, maxOrder);
  if( ! plot) return noPlot();
  Points * lengthPoints = plot->getNewPoints( getID(), "red" );
  int minimumOrder(0), minimumLength(-1);
  double orderStep = maxOrder / stepNumber;
  if(orderStep < 1) orderStep = 1;
  EnhancedProgressBar progress( "two part MDL", maxOrder, 0 );
  if( ! sample->getVanDerMonde()->leastSquare( maxOrder, & progress ) )
    return noMatrix();

  for(double myOrder = 0; myOrder <= maxOrder; myOrder += orderStep){
    setOrder( int(myOrder) );
    if ( progress.check(order) ) break;
    sample->getVanDerMonde()->makeParameter( parameter, precisionArray, order );
    int length = goodnessOfFit(variance);
    if( length < minimumLength || minimumLength < 0 ){
      minimumLength = length;
      bottom = length + 2 ;
      minimumOrder = order;
    }
    if( bottom > length ){
      bottom = length;
      plot->getAxes()->setLowerY( bottom -1 );
    }
    if( top < length ){
      top = length;
      plot->getAxes()->setUpperY( top + 1);
    }
    lengthPoints->paint_back((double)order, double(length));
    plot->refresh();
  }
  double lowerLimit = bottom - (top - bottom) / 10;
  plot->getAxes()->setLowerY( lowerLimit < 0 ? 0 : lowerLimit );
  plot->refresh();
  setOrder( minimumOrder );
//  sample->getVanDerMonde()->makeParameter( parameter, precisionArray, order );
  return true;
}

bool LearnPolynomial::orderPrecision(Sample * sample){// return true; }/*
  ScientificCanvas * plot = makePlot(Scale::SIMPLE, 0, maxOrder, 0, maxOrder);
  if( ! plot) return noPlot();
  Points * lengthPoints = plot->getNewPoints( getID(), "red" );
  int minimumOrder(0), minimumPrecision(0), minimumLength(-1);
  int length[ maxOrder + 1 ];
  double step = maxOrder / stepNumber;
  if(step < 1) step = 1;
  EnhancedProgressBar progress( "order - precision", maxOrder, 0 );
  if( ! sample->getVanDerMonde()->leastSquare( maxOrder, & progress ) )
    return noMatrix();

  for(double myOrder = 0; myOrder <= maxOrder; myOrder += step){
    order = int(myOrder);
    if ( progress.check(order) ) break;
    int myMinimumPrecision(0), myMinimumLength(-1);

    for(double myPrecision = step; myPrecision <= maxOrder; myPrecision += step){
      if ( progress.check() ){ myOrder = maxOrder + 1; break; }
      setOrder( order, (int)myPrecision );
      sample->getVanDerMonde()->makeParameter( parameter, precisionArray, order );
      length[precision] = goodnessOfFit( variance );
//      cerr << "order " << order << " \t precision " << parameterPrecision
//           << " \t length " << length[parameterPrecision] << endl;
      if( length[precision] < myMinimumLength || myMinimumLength < 0 ){
        myMinimumLength = length[precision];
        myMinimumPrecision = precision;
      }
    }
    if(myMinimumLength < minimumLength || minimumLength < 0){
      minimumLength = myMinimumLength;
      minimumPrecision = myMinimumPrecision;
      minimumOrder = order;
    }
    Points * precisionPoints = plot->getNewPoints( getID(), "green" );
    for(double myPrecision = step; myPrecision <= maxOrder; myPrecision += step){
      double myLength = order + step
                        * ( (double)myMinimumLength / length[int(myPrecision)] - 1);
      precisionPoints->paint_back(myLength, double(int(myPrecision)));
    }
    lengthPoints->paint_back((double)order, myMinimumPrecision);
    plot->refresh();
  }
  setOrder( minimumOrder, minimumPrecision );
  return true;
}
//=================================================================================================
bool LearnPolynomial::crossValidate(Sample * sample){
  VanDerMonde * learnVanDerMonde = sample->getVanDerMonde();
  Points * learnPoints = sample->getPoints();
  int count = learnPoints->size();
  double bottom(1), top(10);
  ScientificCanvas * plot = makePlot(Scale::LOG,
                                     bottom, top, 0, maxOrder);
  if( ! plot ) return noPlot();
  int fractions = 10;
  EnhancedProgressBar progress( "cross validation", fractions, 0);
  int index[count];
  for(int i=0; i<count; i++) index[i] = i;
  random_shuffle(index, index + (count-1) );
  Points group[fractions];
  VanDerMonde matrixes[fractions];
  for(int i=0; i<count; i++)
    group[i % fractions].copy(( * learnPoints )[index[i]]);
  int frac;
  for(frac=0; frac<fractions; frac++){
    matrixes[frac].setPoints( & (group[frac]) );
    matrixes[frac].setOrder(maxOrder);
    if( ! matrixes[frac].processData( & progress ) )
      return xml_warning("can't process auxiliary set");
  }
  if( learnVanDerMonde->getOrder() < maxOrder ){
    learnVanDerMonde->setOrder( maxOrder );
    for(frac=0; frac<fractions; frac++) (*learnVanDerMonde) += matrixes[frac];
  }

//--------------------------------   work on mother for each group

  VanDerMonde mother;
  mother.setOrder(maxOrder);
  double sum[maxOrder + 1];
  for(int i=0; i<=maxOrder; i++) sum[i]=0;
  int step = maxOrder / stepNumber;
  if( step < 1 ) step = 1;
  int hue = 175, saturation = 250, hsv_value = 150;
  bool stop = false;
  for(frac=0; frac<fractions; frac++){
    if ( stop = progress.check(frac) ) break;
    mother.makeZero();
    for(int j=0; j<fractions; j++) if(j != frac)  mother += matrixes[j];
    //    for(int k=0; k<mother.sumX.getSize(); k++){ cerr << " mother.sumX[" << k << "] = " << Matrix::toDouble(mother.sumX[k]) << endl; }
    mother.makeTriangular( & progress );
    Points * plotPoints = plot->getNewPoints( getID(), QColor(hue, saturation, hsv_value, QColor::Hsv).name());
    hsv_value += 10;                //    cerr << " best fit " << frac << " : " << mother.bestFit(&(group[frac])) << " " << endl;
    for(int degree=0; degree <=maxOrder; degree += step){
      if ( stop = progress.check() ) break;
      setOrder( degree );
      mother.makeParameter( parameter, precisionArray, degree );
      double result = matrixes[frac].variance(parameter, degree);    //  mother.variance(&(matrixes[frac]));
      sum[degree] += result;        //     cerr << " newPoint \tx " << degree << " \ty result " << result << endl;
      plotPoints->paint_back((double)degree, result);
      if( result > top ) {
        top = result;
        plot->getAxes()->setUpperY(top);
      } else if( result < bottom ){
        bottom = result;
        plot->getAxes()->setLowerY(bottom);
      }
      plot->refresh();
    }
    if(stop) break;
  }

  Points * totalPoints = plot->getNewPoints( getID(), "red" );
  smooth(sum, maxOrder + 1, step);
  double minimum=1000000000;
  for(int degree = 0; degree <= maxOrder; degree += step){
    sum[degree] /= frac;
    if(sum[degree] < minimum){
      minimum = sum[degree];
      order = degree;
    }
    totalPoints->push_back(double(degree), sum[degree]);
  }
  totalPoints->recalculate();
  plot->refresh();
  setOrder(order);
  return true;
}
void LearnPolynomial::smooth(double * array, const int size, const int step){
  double last = 0;
  int degree=0;
  for(degree = 0; degree < size-step; degree += step){
    double last2 = array[degree];
    array[degree] += last + array[degree + step];
    last = last2;
  }
  array[degree] += last;
  for(int degree = 0; degree < size; degree++){
    if(degree != 0 && degree < size - step) array[degree] /= 3;
    else array[degree] /= 2;
  }
}
bool LearnPolynomial::noTest(){
  return xml_warning("no test sample selected");
}
bool LearnPolynomial::noPlot(){
  return xml_warning("can't make extra plot for graph");
}
bool LearnPolynomial::noMatrix(){
  return xml_warning("can't make matrix for learn sample");
}
bool LearnPolynomial::emptyTest(){
  return xml_warning("test sample is empty");
}
bool LearnPolynomial::noTestProcess(){
  return xml_warning("can't process test sample");
}
