/***************************************************************************
                          matrix.cpp  -  description
                             -------------------
    begin                : ?' ??? 5 2002
    copyright            : (C) 2002 by Volker Nannen
    email                : volker@nannen.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "matrix.h"
#include "enhancedprogressbar.h"

#include <m_apm.h>

#include <qdom.h>

#include <cstdlib>

using namespace std;

Matrix::Matrix(const QString name)
  : BasicElement(0, name),
    array(0),
    size(0),
    precision(0)
{
  init(1,1);
}
Matrix::Matrix(const int x, const int y)
  : BasicElement(0),
    array(0),
    size(0),
    precision(0)
{
  init(x, y);
}
Matrix::Matrix(Matrix & other)
  : BasicElement(0),
    array(new MAPM[other.size]),
    size(other.size),
    rows(other.rows),
    columns(other.columns),
    precision(other.precision)
{
  for(int i=0; i<size; i++) array[i] = other.array[i];
}
Matrix::~Matrix(){
  destroy();
}
bool Matrix::xml_read(QDomElement * element){
  BasicElement::xml_read(element);
  init(rows, columns);
  QDomNodeList list = element->elementsByTagName("MAPM");
  int max = (int)list.count() < size ? list.count() : size;
  EnhancedProgressBar progress( "reading matrix ...", max);
  for(int i=0; i<max; i++){
    if( progress.check( 63, i ) ) break;
    QDomElement MAPM_Element = list.item(i).toElement();
    int index = MAPM_Element.attribute("index").toInt();
    array[index] = MAPM_Element.attribute("value", "0");
  }
  return true;
}
bool Matrix::xml_write(QDomElement * parent){
  BasicElement::xml_write( parent );
  QDomElement * thisElement = thisDomElement();
  char out[precision + 20];
  for(int i = 0; i < size; i++){
    QDomElement MAPM_Element = thisElement->ownerDocument().createElement("MAPM");
    MAPM_Element.setAttribute("index", i);
    if( array[i].sign() == 0 ){
      MAPM_Element.setAttribute("value", QString("0"));
    } else {
      array[i].round(precision);
      array[i].toString(out, precision+2);
      MAPM_Element.setAttribute("value", QString(out));
    }
    thisElement->appendChild(MAPM_Element);
  }
  return true;
}
void Matrix::xml_list(){
  BasicElement::xml_list();
  xml_int("size", size, 1);
  xml_int("rows", rows, 1);
  xml_int("columns", columns, 1);
  xml_int("precision", precision, 1);
}
void Matrix::init(int x, int y){
  if(x < 0) x=0;
  if(y < 0) y=0;
  int newSize = x*y;
  size = newSize;
  rows = x;
  columns = y;
  if(array) delete[] array;
  array = (size ? new MAPM[size] : 0);
  makeZero();
}
void Matrix::resize(int x, int y){
  if(x < 0) x=0;
  if(y < 0) y=0;
  int newSize = x*y;
  MAPM * tmpArray = new MAPM[newSize];
  for(int i=0; i<newSize; i++) tmpArray[i] = MM_Zero;
  for(int row = 0; row < rows && row < x; row++){
    for(int column = 0; column < columns && column < y; column++){
      tmpArray[row*y + column] = array[row*columns + column];
    }
  }
  size = newSize;
  rows = x;
  columns = y;
  if(array) delete[] array;
  array = tmpArray;
}
void Matrix::destroy(){
  if(array){
    delete[] array;
    array = 0;
    size = 0;
    columns = 0;
    rows = 0;
  }
}
Matrix & Matrix::operator+=(Matrix & other){
  for(int i=0; i<size && i<other.size; i++){
    array[i] += other.array[i];
    array[i].round(precision);
  }
  return *this;
}
void Matrix::swapRows(const int & a, const int & b){
  MAPM tmp;
  for(int i = 0; i < columns; i++){
    tmp = array[columns*a + i];
    array[columns*a + i] = array[columns*b + i];
    array[columns*b + i] = tmp;
  }
}
void Matrix::swap(const int & a, const int & b){
  MAPM tmp = array[a];
  array[a] = array[b];
  array[b] = tmp;
}
void Matrix::makeZero(){
  if(array){
    for(register int i = 0; i<size; i++){
      array[i] = MM_Zero;
    }
  }
}
bool Matrix::makeTriangular(EnhancedProgressBar * dialog){
  int dialogLimit = (rows) * (rows) * (rows);
  EnhancedProgressBar progress(dialog, "least square fitting...", dialogLimit);
  for(int pivot = 0; pivot < rows; pivot++){
    if( progress.check(dialogLimit - (rows -pivot) * (rows - pivot)  *
                                     (rows - pivot)) )
      return false;
    if( matrix( pivot, pivot ).sign() <= 0 ) {
      cerr << " matrix singular from [" << pivot << "," << pivot << "]\n";
      for( int row = pivot; row < rows; row++){
        for(int col = pivot; col < columns; col++){
          matrix( row, col ) = MM_Zero;
        }
      }
      break;
    }
    for(int col = pivot + 1; col < columns; col++){
                 // implicitly matrix(pivot, pivot) = 1
      matrix(pivot, col) /= matrix(pivot, pivot);
      matrix(pivot, col).round(precision );
    }
    for(int row = pivot + 1; row < rows; row++){
      for(int col = pivot + 1; col < columns; col++){
        matrix(row, col) -= ( matrix(row, pivot) * matrix(pivot, col) ).round(precision );
        matrix(row, col).round(precision );
      }
      matrix(row, pivot) = MM_Zero;
    }
  }
  return true;
}
double Matrix::determinant(int degree){
  if(degree < 0) degree = 0;
  if(degree >= rows) degree = rows-1;
  MAPM det=1;
  for(int i=0; i<=degree; i++){
    det *= matrix(i, i);
    det.round(precision);
  }
  return toDouble(det);
}
double Matrix::logDeterminant(int degree, bool & valid){
  if( degree == 1){
    char out[100];
    for(int i=0; i<rows; i++){
      matrix(i,i).toString(out, 3);
      cerr << " [" << i << "]: " << out;
      if( ! (i % 10) ) cerr << endl;
    }
  }
  if( degree < 0 ) degree = 0;
  if( degree > rows) degree = rows;
  MAPM det=1;
  for(int i=0; i<=degree; i++){
    det *= matrix(i, i);
    det.round(precision);
  }
  if( det.sign() == 0 ) {
    valid = false;
    cerr << " warning : singular matrix " << toDouble(det) << endl;
    return -1e+300;
  }
  return toDouble(log(det, precision));
}
double Matrix::toDouble(const MAPM & value){
  static char doubleBuffer[200];
  value.toString(doubleBuffer,10);
  return atof(doubleBuffer);
}
QString Matrix::toString(const int index){
  if(index < 0 || index >= size) return QString::null;
  char out[precision + 100];
  array[index].toString(out, -1);
  return QString(out);
}

