/***************************************************************************
                          plot.cpp  -  description
                             -------------------
    begin                : Fri Mar 1 2002
    copyright            : (C) 2002 by Volker Nannen
    email                : volker@nannen.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "plot.h"
#include "project.h"
#include "axes.h"
#include "sdvstatusbar.h"
#include "workspace.h"
#include "icons/crosshair.xpm"

#include <qfiledialog.h>
#include <qstring.h>
#include <qcolor.h>
#include <qlabel.h>
#include <qpainter.h>
#include <qcursor.h>
#include <qevent.h>

Plot::Plot(Project * project, ScientificCanvas * canvas, const char *name, WFlags f)
  : QCanvasView(0,project->getWorkspace(), name, f),
    statusBar(project->getWorkspace()->getStatusBar()),
    labelX(statusBar->getLabelX()),
    labelY(statusBar->getLabelY()),
    scalePlot(0),
    project(project),
    scaling(false),
    resized(false),
    scaleSize(300),
    scale(3)
{
  setCanvas(canvas);
  viewport()->setMouseTracking(true);
  viewport()->setCursor(QCursor(QPixmap(crosshair),16,16)); // Qt::crossCursor));
  setResizePolicy(QScrollView::AutoOne);
}
Plot::~Plot(){
  destroy();
}
void Plot::destroy(){
  if(scalePlot){
    delete scalePlot;
    scalePlot = 0;
  }
}
void Plot::resizeContents(int x, int h){
//  cerr << " plot resize " << x << " x " << h << " \n";
  QCanvasView::resizeContents(x, h);
  QCanvasView::resize(x + width() - visibleWidth(), h + height() - visibleHeight() );
  viewport()->setBackgroundMode(PaletteDark);
  viewport()->repaint();
}
void Plot::viewportResizeEvent(QResizeEvent * event){
  QCanvasView::viewportResizeEvent(event);
  scientificCanvas->setSize(event->size().width(), event->size().height());
  resized = true;
}
void Plot::closeEvent(QCloseEvent * event){
  if( scientificCanvas && scientificCanvas->close() )
    QCanvasView::closeEvent(event);
}
void Plot::contentsMousePressEvent(QMouseEvent * event){
  if( ! project ) return;
  if( event->button() & QMouseEvent::LeftButton ){
    QCanvasItemList list = canvas()->collisions(event->pos());
    if( ! list.isEmpty() ){
      project->selectElement( dynamic_cast<BasicItem * >( list.first() )->getID() );
    }
  }
  if( event->button() & QMouseEvent::MidButton ){
    if( ! scalePlot ){
      scalePlot = new ScalePlot(this->viewport(), scale, scaleSize, "scale plot");
      scalePlot->setCanvas(canvas());
//      scalePlot->resizeContents( int((canvas()->width()+scaleSize)*scale), int((canvas()->height()+scaleSize)*scale));
    }
    scaling = true;
    scalePlot->move(event->pos().x(), event->pos().y(), contentsX(), contentsY());
    scalePlot->show();
  }
  project->mousePressed(event);
}
void Plot::contentsMouseDoubleClickEvent(QMouseEvent * event){
  if( ! project ) return;
  project->mouseDoubleClicked(event);
}
void Plot::contentsMouseMoveEvent(QMouseEvent * event){
//  cerr << " posX " << event->pos().x() << " posY " << event->pos().y() << endl;
  if( ! project ) return;
  if( axes->posToVal(event->x(), event->y(), currentX, currentY) ){
    labelX->setNum(currentX);
    labelY->setNum(currentY);
  } else {
    labelX->setText("-");
    labelY->setText("-");
  }
  if(scaling){
//    if( event->button() & QMouseEvent::MidButton ){
      scalePlot->move(event->pos().x(), event->pos().y(), contentsX(), contentsY());
//    } else {
//      scalePlot->hide();
//      scaling = false;
//    }
  }
  QCanvasItemList list = canvas()->collisions(event->pos());
  if ( ! list.isEmpty() ){
//    cerr << " found " << dynamic_cast<BAsicElement * >( list.first() )->getElementName() << endl;
    project->highlightElement( dynamic_cast<BasicItem * >( list.first() )->getID() );
   } else {
    project->highlightElement(0);
  }
  project->mouseMoved(event);
}
void Plot::contentsMouseReleaseEvent(QMouseEvent * event){
  if( ! project ) return;
  if(scaling && (event->button() & QMouseEvent::MidButton) ){
    scalePlot->hide();
    scaling = false;
  }
  project->highlightElement(0);
  project->mouseReleased(event);
}
void Plot::mouseReleaseEvent(QMouseEvent * /*event*/){
  if(resized){
    resized = false;
    scientificCanvas->refresh();
  }
}
void Plot::leaveEvent(QEvent * /*event*/){
  if(resized){
    resized = false;
    scientificCanvas->refresh();
  }
}
void Plot::enterEvent(QEvent * /*event*/){
  if(resized){
    resized = false;
    scientificCanvas->refresh();
  }
}
void Plot::setCanvas(ScientificCanvas * newCanvas){
  scientificCanvas = newCanvas;
  if(newCanvas){
    QCanvasView::setCanvas(newCanvas);
    viewport()->setBackgroundMode(PaletteDark);
    viewport()->repaint();
    if(scalePlot){
      scalePlot->setCanvas(newCanvas);
    }
    axes = newCanvas->getAxes();
  } else {
    QCanvasView::setCanvas(0);
    axes = 0;
  }
}
void Plot::save(){
  cerr << " saving plot \n";
  QString fileName = project->getPlotFileName();
  if ( fileName.isNull() ) return;
  QPixmap pixmap( canvas()->size() );
  QPainter *painter = new QPainter( &pixmap );
  drawContents( painter, 0, 0, width(), height() );
  painter->end();
  pixmap.save( fileName, "PNG" );
}

//------------------------------------------------------class ScalePlot

ScalePlot::ScalePlot(QWidget *parent, double scale, int size, const char *name)
  : QCanvasView(0,parent,name, WStyle_Customize | WPaintUnclipped \
                    | WPaintClever | WRepaintNoErase),
    parent(parent),
    scale(scale),
    scaleSize(size)
{
  viewport()->setMouseTracking(true);
  setHScrollBarMode(AlwaysOff);
  setVScrollBarMode(AlwaysOff);
  setFocusProxy(parent);
  setFixedWidth(scaleSize);
  setFixedHeight(scaleSize);
  offset = int(scaleSize/(2*scale));
}
void ScalePlot::move(int x, int y, int contX, int contY){
  QCanvasView::move(x - (scaleSize >> 1) - contX, y - (scaleSize >> 1) - contY);
  setContentsPos(int(x*scale) + 3, int(y*scale) + 3);
  repaint(x - offset, y - offset, 2*offset, 2*offset);
  viewport()->update();
}
void ScalePlot::drawContents( QPainter * p, int cx, int cy, int cw, int ch ){
//  p->fillRect(cx, cy, cw, ch, backgroundColor());//, bgColor);
  p->scale(scale, scale);
  p->translate(offset, offset);
  QCanvasView::drawContents ( p, int(cx/scale - offset), int(cy/scale - offset),
                                 int(cw/scale), int(ch/scale));
}
void ScalePlot::resizeContents(int w, int h){
  QCanvasView::resizeContents(int(w*scale)+2*scaleSize, int(h*scale)+2*scaleSize);
  viewport()->setBackgroundMode(PaletteDark);
  viewport()->repaint();
}
