/***************************************************************************
                          points.cpp  -  description
                             -------------------
    begin                : Thu May 9 2002
    copyright            : (C) 2002 by Volker Nannen
    email                : volker@nannen.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "points.h"
#include "singlepoint.h"
#include "axes.h"
#include "enhancedprogressbar.h"

#include <qdom.h>

Points::Points(const int ID, ScientificCanvas * canvas)
  : vector<SinglePoint *>(),
    BasicElement(ID, "graph"),
    canvas(canvas),
    continuous(false),
    visible(true),
    thick(false),
    thickness(0),
    outside(0)
{
  hideName();
}
Points::Points(Points & other, const int ID, ScientificCanvas * canvas)
  : vector<SinglePoint *>(),
    BasicElement(ID),  // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!1
    canvas(canvas),
    continuous(other.continuous),
    visible(other.visible),
    thick(other.thick),
    thickness(other.thickness),
    outside(0),
    color(other.color),
    defaultColor(other.defaultColor),
    pen(other.pen)
{
  makeControl();
  hideName();
  copy(other);
}
Points::~Points(){
  destroy();
}
void Points::clear(){
  destroy();
}
void Points::destroy(){
  if(size()){
    for(Points::iterator iter = begin(); iter != end(); iter++){
      delete (*iter);
    }
    vector<SinglePoint *>::clear();
  }
}
bool Points::copy(Points & other, EnhancedProgressBar * dialog){
  if(other.size()){
    int count = 0;
    EnhancedProgressBar progress( dialog, "Copying points ...", other.size());
    for(Points::iterator iter = other.begin(); iter != other.end(); iter++, count++){
      if ( progress.check(63, count) ) return false;
      copy(*iter);
    }
  }
  return true;
}
bool Points::xml_read(QDomElement * element, EnhancedProgressBar * dialog){
  destroy();
  BasicElement::xml_read(element);
  setContinuous(continuous, dialog);
  setVisible(visible);
  setThick(thick);
  setColor(color);
  makeControl();
  if(element){
    QDomNodeList pointList = element->elementsByTagName("Point");
    QDomElement pointElement;
    if( pointList.length() ){
      EnhancedProgressBar progress(dialog, "loading points ...", pointList.length());
   //   cerr << pointList.length() << " points from dom\n";
      for(unsigned int i = 0; i < pointList.length(); i++){
        if ( progress.check(63, i) )   break;
        pointElement = pointList.item(i).toElement();
        push_back( & pointElement );
      }
    }
  }
  return true;
}
bool Points::xml_write(QDomElement * parent){
  BasicElement::xml_write( parent );
  if(size()){
//  if(size() && preserve){
    for(Points::iterator item = begin(); item != end(); item++){
      (*item)->xml_write( thisDomElement() );
    }
  }
  return true;
}
void Points::xml_list(){
  BasicElement::xml_list();
  xml_bool("thick", thick, false, EDIT);
  xml_bool("continuous", continuous, false, EDIT);
  xml_bool("visible", visible, true, EDIT);
  xml_qstring("color", color, defaultColor, EDIT);
}
void Points::makeControl(){
  controlVisible    = visible;
  controlContinuous = continuous;
  controlColor  = color;
  controlThick = thick;
}
void Points::push_back(SinglePoint * newPoint){;
  newPoint->setPen(pen);
  newPoint->setVisible(visible);
  if(continuous && (size() > 0)){
    SinglePoint * previous = back();
    newPoint->setPoints(previous->getPositionX(), previous->getPositionY(),
                        newPoint->getPositionX(), newPoint->getPositionY() );
  }
  vector<SinglePoint * >::push_back(newPoint);
}
void Points::paint_back(const double & valX, const double & valY, const bool & valid){
  int posX, posY;
  if(canvas) canvas->getAxes()->valToPos(valX, valY, posX, posY);
  push_back(new SinglePoint(getID(), canvas, posX, posY, valX, valY, thickness, valid));
}
void Points::setContinuous(bool yes, EnhancedProgressBar * dialog){
  continuous = yes;
  controlContinuous = yes;
  if( empty() ) return;
  int count = 0;
  EnhancedProgressBar progress(dialog, "drawing points ...", size());
  if(continuous){
    Points::iterator iter = begin();
    int oldX = (*iter)->getPositionX();
    int oldY = (*iter)->getPositionY();
    (*iter)->makeLine(oldX, oldY);
    for(iter++; iter != end(); iter++, count++){
      if ( progress.check(63, count) )  break;
      (*iter)->makeLine(oldX, oldY);
    }
  } else {
    for(Points::iterator iter = begin(); iter != end(); iter++, count++){
      if ( progress.check(63, count) )  break;
      (*iter)->makePoint();
    }
  }
  setVisible(visible);
  if(canvas) canvas->update();
}
void Points::setPen(const QPen & newPen){
  pen = newPen;
  for(Points::iterator iter = begin(); iter != end(); iter++){
    (*iter)->setPen(pen);
  }
  if(canvas && size()) canvas->update();
}
void Points::setThick(const bool yes){
  thick = yes;
  controlThick = thick;
  thickness = thick ? 1 : 0;
  for(Points::iterator iter = begin(); iter != end(); iter++){
    (*iter)->setThickness(thickness);
  }
  if(canvas && size()) canvas->update();
}
void Points::setVisible(const bool yes){
  visible = yes;
  controlVisible = yes;
  if(continuous){
    SinglePoint * previous, * current;
    bool prepreValid, preValid, currentValid;
    if(size()){
      previous = (*this)[0];
      preValid = previous->getValid();
      previous->setVisible( yes && preValid );
      if(size() > 1){
        current = (*this)[1];
        currentValid = current->getValid();
        for(unsigned int i = 2; i<size(); i++){
          previous = current;
          current = (*this)[i];
          prepreValid = preValid;
          preValid = currentValid;
          currentValid = current->getValid();
          previous->setVisible(yes && (prepreValid || preValid || currentValid) );
        }
        current->setVisible(yes && (preValid || currentValid) );
      }
    }
  } else {
    for(Points::iterator iter = begin(); iter != end(); iter++){
      (*iter)->setVisible(yes && (*iter)->getValid());
    }
  }
  if(canvas && size()) canvas->update();
}
void Points::setDefaultColor(const QString newColor){
  defaultColor = newColor;
  color = newColor;
  controlColor = color;
  setPen(QPen(QColor(color)));
}
void Points::setCanvas(ScientificCanvas * newCanvas){
  canvas = newCanvas;
  for(Points::iterator iter = begin(); iter != end(); iter++){
    (*iter)->setCanvas(canvas);
  }
}
void Points::setID(const int newID){
  BasicElement::setID(newID);
  for(Points::iterator iter = begin(); iter != end(); iter++){
    (*iter)->setID(newID);
  }
}
int Points::recalculate(EnhancedProgressBar * dialog){
  outside = 0;
  int count = 0;
  if( size() && canvas) {
    Axes * axes = canvas->getAxes();
    EnhancedProgressBar progress(dialog, "recalculating points ...", size());
    for(Points::iterator iter = begin(); iter != end(); iter++, count++){
      if ( progress.check(63, count) ) break;
      if( ! (*iter)->recalculate(axes)) outside++;
    }
    setContinuous(continuous);
//    setContinuous(continuous, & progress);
  }
  return outside;
}
void Points::bounds(double & up, double & down, double & left, double & right){
  if( ! size() ) return;
  up    = (*begin())->getValueY();
  down  = (*begin())->getValueY();
  left  = (*begin())->getValueX();
  right = (*begin())->getValueX();
  int count = 0;
  EnhancedProgressBar progress( "calculating maximum values", size());
  for(Points::iterator iter = begin(); iter != end(); iter++, count++){
    if ( progress.check(63, count) )  break;
    if( up    < (*iter)->getValueY() ) up    = (*iter)->getValueY();
    if( down  > (*iter)->getValueY() ) down  = (*iter)->getValueY();
    if( left  > (*iter)->getValueX() ) left  = (*iter)->getValueX();
    if( right < (*iter)->getValueX() ) right = (*iter)->getValueX();
  }
}
void Points::setColor(const QString newColor){
  color = newColor;
  controlColor = color;
  setPen(QPen(QColor(color)));
}
void Points::refresh(){
  if(controlVisible != visible){
    setVisible(visible);
  }
  if(controlContinuous != continuous){
    setContinuous(continuous);
  }
  if(controlColor != color){
    setColor(color);
  }
  if(controlThick != thick){
    setThick(thick);
    setContinuous(continuous);
  }
}
