/***************************************************************************
                          polynomial.cpp  -  description
                             -------------------
    begin                : Thu May 2 2002
    copyright            : (C) 2002 by Volker Nannen
    email                : volker@nannen.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "polynomial.h"
#include "project.h"
#include "sample.h"
#include "collection.h"
#include "scientificcanvas.h"
#include "axes.h"
#include "precisioneditor.h"
#include "parsenode.h"
#include "vandermonde.h"
#include "enhancedprogressbar.h"
//#include "selectionmethod.h"

#include <algorithm>
#include <cmath>

using namespace std;

Polynomial::MethodEnumeration Polynomial::methodEnumeration;

Polynomial::MethodEnumeration::MethodEnumeration()
  : DomEnumeration("sample method enumeration")
{
  insert("manual", MANUAL);
//  insert("trimmed MDL", TRIMMED_MDL);
  insert("two part MDL", TWO_PART_MDL);
  insert("mixture MDL", MIXTURE_MDL);
  insert("cross validation", CROSS);
  insert("analysis", ANALYSIS);
  insert("complex analysis", COMPLEX);
//  insert("order-precision", ORDER_PRECISION);
}

Polynomial::Polynomial(Project * project, const QString name,
                       QDomElement * element)
  : Graph(project, name),
    penaltyTerm( new ParseNode() ),
    precisionArray( new int[1] )
{
  init();
  xml_read(element);
  refresh();
}
Polynomial::Polynomial(Polynomial * other)
  : Graph(other->project),
    precisionArray( new int[1] )
{
  init();
  xml_copy(other);
  setElementName(project->makeName(getElementName()));
  refresh();
}
void Polynomial::init(){
  parameter.setElementName("parameter");
  points.setDefaultColor("darkgreen");
  graphEnumeration = new DomEnumeration("graph enumeration");
  canCopy = true;
  penaltyTerm->parse(penalty);
}
Polynomial::~Polynomial(){
  destroy();
  if(graphEnumeration) { delete graphEnumeration; /*graphEnumeration = 0;*/ }
  delete[] precisionArray;
  precisionArray = 0;
}
void Polynomial::destroy(){
  Graph::destroy();
  parameter.makeZero();
  variance = 0;
  testVariance = 0;
//  if (selectionMethod) delete selectionMethod;
//  selectionMethod=0;
}
bool Polynomial::xml_read(QDomElement * element){
  destroy();
  points.setContinuous(true);
  Graph::xml_read(element);
  makeControl();
  if(element){
    QDomNodeList list = element->elementsByTagName("Matrix");
    makeXML_Element( & parameter, & list);
  }
  penaltyTerm->parse(penalty);
  setMaxOrder(maxOrder);
  setOrder(order, precision);
  return true;
}
void Polynomial::makeControl(){
  controlMaxOrder = maxOrder;
  controlOrder = order;
  controlPrecision = precision;
  controlLearnSampleID = learnSampleID;
  controlTestSampleID = testSampleID;
  controlMethod = method;
  controlFit = fit;
  controlPenalty = penalty;
}
void Polynomial::xml_list(){
  graphEnumeration->init(getParentCollection(), "Sample");
  graphEnumeration->add(getParentCollection(), "Merge");
  Graph::xml_list();
  xml_bool("valid", valid, true, HIDE);
  xml_double("residual test noise", testVariance, 0);
  xml_select(graphEnumeration, "test", testSampleID, -1);
  xml_double("residual noise", variance, 0);
  xml_int("description length", descriptionLength, 0);
  xml_bool("edit precision", editPrecision, false, EDIT);
  xml_qstring("penalty", penalty, "n*log s + k*log n",
   (method == MANUAL || method == TRIMMED_MDL
   || method == TWO_PART_MDL ? EDIT : HIDE) );
  xml_int("number of samples", sampleNumber, 10,
   (method == COMPLEX ? EDIT: HIDE) );
  xml_int("number of steps", stepNumber, 1000,
   (method == TRIMMED_MDL || method == MANUAL ? HIDE : EDIT) );
  xml_int("precision", precision, 0,
   (method == MANUAL || method == TRIMMED_MDL ? EDIT : HIDE) );
  xml_int("order", order, 0,
   (method == MANUAL || method == TRIMMED_MDL ? EDIT : SHOW) );
  xml_int("maximum order", maxOrder, 50,
    (method == MANUAL || method == TRIMMED_MDL ? HIDE : EDIT) );
  xml_select(graphEnumeration, "learn", learnSampleID, -1);
  xml_enum( & methodEnumeration, "method", method, MANUAL);
  xml_bool("fit", fit, false, EDIT);
  xml_object( & parameter, HIDE);
//  xml_object(selectionMethod, HIDE);
}
void Polynomial::setID(const int ID){
  Graph::setID(ID);
//  if( selectionMethod ) selectionMethod->setID(ID);
}
void Polynomial::setValid(const bool yes){
//  cerr << " polynomial set valid " << (yes?"true" : "false");
  Graph::setValid(yes);
}
void Polynomial::recalculate(){
  valid = true;
  points.clear();
  if( ! controlFit ) return;
  Axes * axes = project->getCanvas()->getAxes();
  if ( ! axes ) return;
  double end = axes->end();

  double step = axes->step();
  int counter(0);
  EnhancedProgressBar progress("drawing polynomial...", int( (end - axes->begin())/step ) );
  for(double xVal = axes->begin(); xVal <= end; xVal += step, counter++){
    if( progress.check( 31, counter ) ) break;
    double invX = axes->invScaleX(xVal);
    double yVal = evaluate(invX);
//    cerr << " inv xVal: " << invX << " yval " << yVal <<endl;
    points.push_back(invX, yVal);
  }
//  points.recalculate( & progress );
  points.recalculate();
  updateCanvas();
  return;
}
//================================================================================================
void Polynomial::refresh(){
  if(controlMethod != method){
    controlMethod = method;
//    makeMethod();
    setElementName(project->makeName(methodEnumeration.stringAtValue(method),
                                     false) );
    xml_edit();
    fit = false;
  }
  if(editPrecision){
    PrecisionEditor(this);
    editPrecision = false;
    emit valueChanged();
  }
  if( controlMaxOrder != maxOrder ){
    setMaxOrder( maxOrder );
    fit = false;
  }
  if(penalty != controlPenalty){
    penaltyTerm->parse(penalty);
    controlPenalty = penalty;
    descriptionLength = goodnessOfFit(variance);
    emit valueChanged();
  }
  if( controlPrecision != precision || controlOrder != order ){
    setOrder( order, precision );
    fit = false;
  }
  if( controlLearnSampleID != learnSampleID ){
    controlLearnSampleID = learnSampleID;
    xml_edit();
    fit = false;
  }
  if(controlTestSampleID != testSampleID){
    controlTestSampleID = testSampleID;
    testGoodness();
    xml_edit();
  }
  if(fit != controlFit){
    if( fit ) fit = makeParameter();
    controlFit = fit;
    if( ! fit ){
      if(points.size()){
        points.clear();
        updateCanvas();
      }
    }
    descriptionLength = goodnessOfFit(variance);
    testGoodness();
    xml_edit();
//    emit valueChanged();
    valid = false;
  }
  Graph::refresh();
}
int Polynomial::goodnessOfFit(double & myVariance, Sample * sample){
  if( ! sample ) sample = getSample(learnSampleID);
  if( ! sample || ! fit ){
    myVariance = 0;
    return 0;
  }
  myVariance = sample->getVanDerMonde()->variance(parameter, order);
  return (int)penaltyTerm->eval( getComplexity(),             order,
                                 sample->getPoints()->size(), myVariance);
}
void Polynomial::testGoodness(){
  Sample * testSample = getSample(testSampleID);
  if( ! testSample || ! fit ) testVariance = 0;
  else testVariance = testSample->getVanDerMonde()->variance(parameter, order);
};
void Polynomial::setMaxOrder(int newMaxOrder){
  if( newMaxOrder < 0 ) newMaxOrder = 0;
  int * myArray = new int[newMaxOrder + 1];
  for(int i=0; i<=maxOrder && i<=newMaxOrder; i++) myArray[i] = precisionArray[i];
  for(int i=maxOrder+1; i<=newMaxOrder; i++) myArray[i] = 0;
  delete[] precisionArray;
  precisionArray = myArray;
  parameter.resize(newMaxOrder+1);
  maxOrder = newMaxOrder;
  controlMaxOrder = maxOrder;
  if( order > maxOrder ) setOrder( maxOrder );
}
void Polynomial::setOrder(const int newOrder, const int newPrecision){
//  cerr << " new order: " << newOrder << endl;
  order = newOrder;
  controlOrder = order;
  precision = newPrecision<0 ? order/2 : newPrecision;
  controlPrecision = precision;
  if( order > maxOrder ) setMaxOrder(order);
  else setPrecision(precision);
}
void Polynomial::setPrecision(const int newPrecision){
  precision = newPrecision;
  controlPrecision = precision;
  for(int i=0; i<= order; i++) precisionArray[i] = precision;
  for(int i=order+1; i<=maxOrder; i++) precisionArray[i]=0;
  parameter.setPrecision(precision);
}
void Polynomial::setPrecision(int max, const int * array){
  if(max < 0) return;
  if(max > maxOrder) setMaxOrder(max);
  for(int i=0; i<=max; i++) precisionArray[i] = array[i];
  for(int i=max+1; i<=maxOrder; i++) precisionArray[i] = 0;
  makeParameter();
  descriptionLength = goodnessOfFit(variance);
  testGoodness();
  emit valueChanged();
  recalculate();
}
bool Polynomial::makeParameter(){
  Sample * sample = getSample(learnSampleID);
  if( sample &&
      sample->getVanDerMonde()->makeParameter(parameter, precisionArray, order) ){
    return true;
  } else {
    parameter.makeZero();
    return false;
  }
}
double Polynomial::evaluate(const double & x){
  int myPrecision = int( order * 1.5 + 30);
  m_apm_cpp_precision(myPrecision);
  MAPM y = parameter[0];
  MAPM power(1);
  for(int i = 1; i<=order; i++){
    power *= x;
    power.round(myPrecision);
    y += (parameter[i] * power).round(myPrecision);
    y.round(myPrecision);
  }
  return Matrix::toDouble(y);
}
double Polynomial::getY(const double x){
  Axes * axes = project->getCanvas()->getAxes();
  if( ! axes ) return 0;
  return evaluate(axes->invScaleX(x));
}
int Polynomial::getComplexity(){
  int complex = 0;
  for(int i=0; i <= maxOrder; i++)
    complex += precisionArray[i] ? int(precisionArray[i] * 3.322) + 8 : 1;
  return complex;
}
Sample * Polynomial::getSample(const int ID){
  Collection * parentCollection = getParentCollection();
  if( ! parentCollection ) return 0;
  return dynamic_cast<Sample *>( parentCollection->IDToElement(ID) );
}
/*void Polynomial::makeMethod(QDomElement * element){
  if(selectionMethod) delete selectionMethod;
  selectionMethod = 0;
  switch(method){
    case TWO_PART_MDL :
      selectionMethod = new  TwoPartMDLMethod( getID(), element);
      break;
    case MIXTURE_MDL :
      selectionMethod = new  MixtureMDLMethod( getID(), element);
      break;
    case TRIMMED_MDL :
      selectionMethod = new  TrimmedMDLMethod( getID(), element);
      break;
    case CROSS :
      selectionMethod = new  CrossValidationMethod( getID(), element);
      break;
    case ANALYSIS :
      selectionMethod = new  AnalysisMethod( getID(), element);
      break;
    case COMPLEX :
      selectionMethod = new  ComplexAnalysisMethod( getID(), element);
      break;
    case ORDER_PRECISION :
      selectionMethod = new  OrderPrecisionMethod( getID(), element);
      break;
  }
}*/
