/***************************************************************************
                          process.cpp  -  description
                             -------------------
    begin                : Fri May 10 2002
    copyright            : (C) 2002 by Volker Nannen
    email                : volker@nannen.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "project.h"
#include "process.h"
#include "scientificcanvas.h"
#include "axes.h"
#include "timeseries.h"
#include "enhancedprogressbar.h"

#include <qstring.h>

#include <cmath>

using namespace std;

Process::SeriesEnumeration Process::seriesEnumeration;

Process::SeriesEnumeration::SeriesEnumeration()
  : DomEnumeration("series enumeration")
{
  insert("---", NONE);
  insert("autoregression", AUTOREGRESSION);
  insert("sinus", SINUS);
//insert("henon", HENON);
  insert("logistic", LOGISTIC);
  insert("lorenz", LORENZ);
  insert("pendulum", PENDULUM);
  insert("step", STEP);
  insert("thom", THOM);
}

Process::Process(Project * project, const QString name, QDomElement * element)
  : Graph(project, name),
    timeSeries(0),
    make(false)
{
  init();
  xml_read(element);
  refresh();
}
Process::Process(Process * other)
  : Graph(other->project),
    timeSeries(0),
    make(false)
{
  init();
  xml_copy(other);
  setElementName(project->makeName(getElementName()));
  refresh();
}
void Process::init(){
  points.setDefaultColor("red");
  points.setContinuous(true);
//  points.setPreserve(true);
  canCopy = true;
}
Process::~Process(){
  destroy();
}
void Process::destroy(){
  Graph::destroy();
  if(timeSeries) delete timeSeries;
  timeSeries = 0;
}
bool Process::xml_read(QDomElement * element){
  destroy();
  Graph::xml_read(element);
  if(element){
    QDomElement myElement = element->elementsByTagName("TimeSeries").item(0).toElement();
    makeSeries( & myElement);
    if(timeSeries){
      int steps = timeSeries->getSteps();
      if( steps <= 0) steps = 1;
      step = (end-begin) / (steps);
      invStep = 1/step;
    }
  }
  controlSeries = series;
  controlBegin = begin;
  controlEnd = end;
  controlMake = make;
  setID( getID() );
  return true;
}
void Process::xml_list(){
  Graph::xml_list();
  xml_double("end", end, 10, EDIT);
  xml_double("begin", begin, 0, EDIT);
  xml_int("seed", seed, 0, ( series != (int)SINUS ? EDIT : HIDE ) );
  xml_enum( & seriesEnumeration, "series", series, (int)NONE,
           SHOW, timeSeries);
  xml_bool("make", make, false, EDIT);
  xml_object(timeSeries, HIDE);
}
void Process::setID(const int ID){
  Graph::setID(ID);
  if( timeSeries ) timeSeries->setID(ID);
}
void Process::refresh(){
//  cerr << " Process refresh\n";
  if(controlSeries != series){
    controlSeries = series;
    makeSeries();
    setElementName(project->makeName(seriesEnumeration.stringAtValue(series),
                                     false) );
    xml_edit();
    make = false;
  }
  if( controlMake != make ){
    controlMake = make;
    if( make ){
      run();
    } else {
      points.clear();
    }
    valid = false;
  }
  Graph::refresh();
}
void Process::makeSeries(QDomElement * element){
//  cerr << " new series\n";
  if(timeSeries) delete timeSeries;
  timeSeries = 0;
  switch(series){
    case (int)AUTOREGRESSION :
      timeSeries = new  AutoSeries( getID(), element);
      break;
    case (int)SINUS :
      timeSeries = new SinusSeries( getID(), element);
      break;
    case (int)HENON :
      timeSeries = new HenonSeries( getID(), element);
      break;
    case (int)LOGISTIC :
      timeSeries = new LogisticSeries( getID(), element);
      break;
    case (int)LORENZ :
      timeSeries = new LorenzSeries( getID(), element);
      break;
    case (int)PENDULUM :
      timeSeries = new PendulumSeries( getID(), element);
      break;
    case (int)STEP :
      timeSeries = new StepSeries( getID(), element);
      break;
    case (int)THOM :
      timeSeries = new ThomSeries( getID(), element);
      break;
  }
}
void Process::run(){
  points.clear();
  if( ! timeSeries ){
    points.push_back(begin, 0.0);
    points.push_back(end, 0.0);
    return;
  }
  timeSeries->start(seed);
  int steps = timeSeries->getSteps();
  if( steps <= 0) steps = 1;
  step = (end-begin) / (steps);
  invStep = 1/step;
  double yVal = timeSeries->next(begin);
  points.push_back(begin, yVal);
//  cerr << " inv xVal: " << project->getCanvas()->getAxes()->invScale(begin) << " yval " << yVal <<endl;
  double min(yVal), max(yVal);
  int count = 0;
  EnhancedProgressBar progress( "making series ", steps, 2000);
  for(double xVal = begin+step; xVal <= end; xVal += step, count++){
    if( ! (count & 63) && progress.check(count) ) break;
    yVal = timeSeries->next(xVal);
//    cerr << " inv xVal: " << project->getCanvas()->getAxes()->invScale(xVal) << " yval " << yVal <<endl;
    points.push_back(xVal, yVal);
    if( max < yVal ) max = yVal;
    if( min > yVal ) min = yVal;
  }
  project->getCanvas()->getAxes()->expand(max, min, begin, end);
}
void Process::recalculate(){
  valid = true;
  emit valueChanged();
  points.recalculate();
  updateCanvas();
}
double Process::getY(const double x){
  if( ! points.size() || x < begin || x > end || end == begin) return 0;
  double myX = (x-begin) * invStep;
  int index = (int)floor(myX);
//  if(index < 5 || index > steps -5) cerr << " index: " << index << " dif: " << x-(int)floor(x) << " x : " << x << endl;
  double y = points[index]->getValueY();
  if( index >= (int)points.size() - 1 ) return y;
  return y + (myX-index) * (points[index+1]->getValueY() - y);
}
