/***************************************************************************
                          project.cpp  -  description
                             -------------------
    begin                : Tue Mar 19 2002
    copyright            : (C) 2002 by Volker Nannen
    email                : volker@nannen.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "macros.h"
#include "project.h"
#include "sample.h"
#include "learnpolynomial.h"
#include "process.h"
#include "split.h"
#include "sdvstatusbar.h"
#include "macros.h"
#include "workspace.h"

#include <qfile.h>
#include <qtextstream.h>
#include <qfiledialog.h>
#include <qregexp.h>
#include <qapplication.h>

Project::Project(Workspace * workspace, const QString name)
  : BasicElement(0, name),
    canvas(0),
    selectedElement(0),
    highlightedElement(0),
    workspace(workspace),
    refreshing(false)
{
}
Project::~Project(){
  destroy();
}
void Project::destroy(){
  if(graphs.size()){
    for(Collection::reverse_iterator item = graphs.rbegin();
        item != graphs.rend(); item++){
      graphs.remove( (*item)->getID() );
    }
    graphs.clear();
    canvas = 0;
  }
}
bool Project::xml_read(QDomElement * element){
  destroy();
  BasicElement::xml_read(element);
  canvas = new ScientificCanvas(this, "project screen");
  canvas->makePrincipal(true);
  canvas->setCaption(getElementName());
  graphs.push_back(canvas);

  if(element){
    QDomNodeList nodeList = element->elementsByTagName("Canvas");
    makeXML_Element( canvas, & nodeList );
    for (unsigned int i=0; i<nodeList.count(); i++){
      QDomElement myElement = nodeList.item(i).toElement();
      if( myElement.attribute("ID", "").toInt() > 1 )
        graphs.push_back( new ScientificCanvas(this, "canvas", & myElement) );
    }
    nodeList = element->elementsByTagName("Sample");
    for (unsigned int i=0; i<nodeList.count(); i++){
      QDomElement myElement = nodeList.item(i).toElement();
      graphs.push_back(new Sample(this, "sample", & myElement));
    }
    nodeList = element->elementsByTagName("Polynomial");
    for (unsigned int i=0; i<nodeList.count(); i++){
      QDomElement myElement = nodeList.item(i).toElement();
      graphs.push_back(new LearnPolynomial(this, "polynomial", & myElement));
    }
    nodeList = element->elementsByTagName("Process");
    for (unsigned int i=0; i<nodeList.count(); i++){
      QDomElement myElement = nodeList.item(i).toElement();
      graphs.push_back(new Process(this, "process", & myElement));
    }
    nodeList = element->elementsByTagName("Split");
    for (unsigned int i=0; i<nodeList.count(); i++){
      QDomElement myElement = nodeList.item(i).toElement();
      graphs.push_back(new Split(this, "split", & myElement));
    }
  }
  refresh();
  updateCanvas();
  return true;
}
void Project::xml_list(){
  BasicElement::xml_list();
  xml_bool("valid", valid, true, HIDE);
  xml_int("selectedElement", selectedElementID, 1, HIDE);
  xml_int("graphCount", graphs.getCount(), 0, HIDE);
  for(Collection::iterator element = (graphs.begin()); element != graphs.end(); element++){
    xml_object(*element);
  }
}
void Project::setValid(const bool yes){
//  if ( ! yes ) cerr << " project not valid \n";
  valid = yes;
  if(graphs.size()){
    Collection::iterator item = graphs.begin();
    item++;  // NOT canvas
    for(; item != graphs.end(); item++){
      (*item)->setValid(yes);
    }
  }
}
void Project::refresh(){
  if(refreshing) return;
  refreshing = true;
  canvas->refresh();
  if(graphs.size()){
    for(Collection::reverse_iterator item = graphs.rbegin(); item != graphs.rend(); item++){
      (*item)->refresh();
    }
  }
  if( elementNameChanged() ){
    canvas->setCaption(getElementName());
    emit projectNameChanged();
    nameChanged(getID());
  }
  refreshing = false;
}

void Project::deleteGraph(const int ID){
  graphs.remove(ID);
  if( ID == selectedElementID ){
    emit fillListBox();
    emit elementSelected( 0 );
  } else {
    emit fillListBox();
    emit elementSelected( selectedElementID );
  }
  updateCanvas();
}
void Project::addGraph(BasicElement * newGraph){
  graphs.push_back(newGraph);
  emit fillListBox();
  highlightElement(0);
}
void Project::addGraph(GraphType type, bool show){
  Graph * newGraph;
  switch( type ){
    case SAMPLE :
      newGraph = new Sample(this, makeName("sample"));
      break;
    case PROCESS :
      newGraph = new Process(this, makeName("process"));
      break;
    case POLYNOMIAL :
      newGraph = new LearnPolynomial(this, makeName("polynomial"));
      break;
    case SPLIT :
      newGraph = new Split(this, makeName("split"));
      break;
    default :
      newGraph = new Graph(this, makeName("graph"));
  }
  int newID = graphs.push_back(newGraph);
  if(show){
    emit fillListBox();
    selectElement(newID);
  }
}
void Project::refreshElement(const int ID){
//  cerr << " id " << ID;
  BasicElement * element = graphs.IDToElement(ID);
  if( element ) element->refresh();
  else if( ID == 0)  refresh();
}
void Project::selectElement(const int ID){
//  cerr << "Project::selectElement, ID=" << ID << endl;
  if( ID != selectedElementID ){
    selectedElementID = ID;
    selectedElement = graphs.IDToElement(ID);
    highlightElement(0);
    emit elementSelected(ID);
  }
}
void Project::highlightElement(const int ID){
  if( highlightedElementID != ID ){
    highlightedElementID = ID;
    if( highlightedElement ) highlightedElement->toggleColor(false);
    highlightedElement = graphs.IDToElement(ID);
    if( highlightedElement ){
      highlightedElement->toggleColor(true);
//      emit elementHighlighted(ID);
    }
  }
  SDVStatusBar * statusBar = workspace->getStatusBar();
  if( ID && highlightedElement ){
    statusBar->message(highlightedElement->getElementName());
  } else {
    statusBar->clear();
  }
}

bool Project::load(){
  if(fileName.isNull()){
    fileName = getElementName() + ".";
    fileName += APPLICATION_FILE_EXTENSION;
  }
  QString extension("*.");
  extension += APPLICATION_FILE_EXTENSION;
  fileName = QFileDialog::getOpenFileName(fileName, extension);
  if (fileName.isNull() ) {
    return false;
  }
  dirName = getDirName(fileName);
  xml_load(fileName);
  return true;
}
bool Project::save(){
  if(fileName.isNull()){
    fileName = getElementName() + ".";
    fileName += APPLICATION_FILE_EXTENSION;
  }

  QString extension("*.");
  extension += APPLICATION_FILE_EXTENSION;
  fileName = QFileDialog::getSaveFileName(fileName, extension);
  if (fileName.isNull() ) {
    return false;
  }
  dirName = getDirName(fileName);
  xml_save(fileName);
  return true;
}

QString Project::getPlotFileName(){
  if( plotDirName.isNull() ) plotDirName = dirName;
//  QFileDialog fileDialog;
  QString plotFileName =
    QFileDialog::getSaveFileName(plotDirName + "/plot.png", "*.png");
  plotDirName = getDirName(plotFileName);
  return plotFileName;
}
QString Project::getDirName(QString fileName){
  int fields = fileName.contains('/');
  QString name = fileName.section('/', 1, fields,
                                  QString::SectionIncludeLeadingSep);
  return name;
}
QString Project::makeName(QString suggestion, const bool alwaysNumber){
  return makeName(suggestion, alwaysNumber, &graphs);
}
QString Project::makeName(QString suggestion, const bool alwaysNumber,
                          vector<BasicElement * > * myGraphs){
  suggestion = suggestion.stripWhiteSpace();
  QRegExp number("<\\s*(\\d+)\\s*>\\s*$");
  QRegExp suggestRegExp("^\\s*" + suggestion, false);
  bool found(false);
  int max(1);
  if(myGraphs->size()){
    for(vector<BasicElement * >::iterator item = myGraphs->begin();
        item != myGraphs->end(); item++){
      QString name = (*item)->getElementName();
      if( suggestRegExp.search(name) != -1 ){
        found = true;
//        cerr << " suggested found : " << suggestRegExp.cap() << " right : "
//             << name.right( name.length() - suggestRegExp.cap().length() )
//             << endl;
        if( number.search( name.right( name.length()
                                       - suggestRegExp.cap().length()
                                      ) ) != -1 ){
//          cerr << " number found " << number.cap() << endl;
          int val = number.cap(1).toInt();
          if(val > max) max = val;
        }
      }
    }
  }
  if(found || alwaysNumber) suggestion = suggestion + " <" + QString::number(max+(found ? 1 : 0)) + ">";
//  cerr << " new suggestion: " << suggestion << endl;
  return(suggestion);
}
