/***************************************************************************
                          ruler.cpp  -  description
                             -------------------
    begin                : Wed Mar 6 2002
    copyright            : (C) 2002 by Volker Nannen
    email                : volker@nannen.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "ruler.h"
#include "data.h"
#include "axes.h"
#include "scientificcanvas.h"

#include <qstring.h>
#include <qdom.h>

#include <cmath>

using namespace std;

#define INV_TANGENS_MAX 10000

double Ruler::deg2rad =0.01745329252;

Ruler::Ruler(const int ID, ScientificCanvas * canvas, const QString name,
             QDomElement * element)
  : BasicElement(ID, name),
    canvas(canvas),
    scale(0),
    mainLine(ID, canvas),
    virtualTangens(&tangens),
    virtualInvTangens(&invTangens),
    vx1(&x1),
    vx2(&x2),
    vy1(&y1),
    vy2(&y2),
    virtualOriginX(&originX),
    virtualOriginY(&originY),
    lineList(0),
    lineListSize(0),
    textList(0),
    textListSize(0)
{
  scale = new Scale(this);
  hideName();
  xml_read(element);
}
Ruler::~Ruler(){
  if(scale) { delete scale; scale = 0; }
  destroy();
}
bool Ruler::xml_read(QDomElement * element){
  destroy();
  bool result = BasicElement::xml_read(element);
  if(element){
    QDomElement tmpElement = element->elementsByTagName("Scale").item(0).toElement();
    scale->xml_read(&tmpElement);
  } else {
    scale->xml_read(0);
  }
  setID(getID());
  return result;
}
void Ruler::setValid(const bool yes){
//  cerr << getElementName() << " putting valid to "
//       << (yes ? "true ":"false") << "\n";
  valid = yes;
}
void Ruler::refresh(){
//  cerr << getElementName() << "::refresh\n";
  setTangens(tangens);
  if( ! valid ){
//    cerr << " ruler not valid\n";
    valid = true;
    canvas->setValid(false);
  } else canvas->update();
}
void Ruler::setID(const int ID){
  BasicElement::setID(ID);
  if(scale) scale->setID(ID);
  if(lineList) for (int i=0; i<lineListSize; i++) lineList[i]->setID(ID);
  if(textList) for (int i=0; i<textListSize; i++) textList[i]->setID(ID);
  mainLine.setID(ID);
}
void Ruler::xml_list(){
  BasicElement::xml_list();
  xml_bool("valid", valid, true, HIDE);
  xml_bool("visible", visible, true, HIDE);
  xml_bool("numbers", numbers, true, HIDE);
  xml_int("margin", margin, MARGIN_DEFAULT, HIDE);
  xml_int("originX", originX, 0, HIDE);
  xml_int("originY", originY, 0, HIDE);
  xml_double("spikeUnit", spikeUnit, 50, EDIT);
  xml_double("tangens", tangens, 0, HIDE);    // must be last,
  xml_object(scale, SHOW, false);
}
bool Ruler::setHorizontal(const bool & yes){
  horizontal = yes;
  if(horizontal){
    virtualOriginX = &originX;
    virtualOriginY = &originY;
    virtualTangens = &tangens;
    virtualInvTangens = &invTangens;
    vx1=&x1;
    vy1=&y1;
    vx2=&x2;
    vy2=&y2;
  } else {
    virtualOriginX = &originY;
    virtualOriginY = &originX;
    virtualTangens = &invTangens;
    virtualInvTangens = &tangens;
    vx1=&y1;
    vy1=&x1;
    vx2=&y2;
    vy2=&x2;
  }
  return horizontal;
}
void Ruler::setVisible(bool yes){
  visible = yes;
  mainLine.setVisible(yes);
  if(lineListSize){
    for(int i=0; i<lineListSize; i++){
      lineList[i]->setVisible(yes);
    }
  }
  if(textListSize){
    for(int i=0; i<textListSize; i++){
      textList[i]->hide();
    }
  }
}
void Ruler::setPen(const QPen & pen){
  if(lineListSize){
    for(int i=0; i<lineListSize; i++){
      lineList[i]->setPen(pen);
    }
  }
  mainLine.setPen(pen);
}
double Ruler::setAngle(const double & angle){
  setTangens(double(tan(double(angle)*deg2rad)));
  emit invTangensChanged( - invTangens);       // beware of loops !!!
  return angle;
}
double Ruler::getAngle() const {
  return double( (atan(tangens)) /deg2rad );
}
double Ruler::smooth(const double & value){
  if (value < -INV_TANGENS_MAX) {
    return -INV_TANGENS_MAX;
  } else if (value > INV_TANGENS_MAX) {
    return INV_TANGENS_MAX;
  } else if (value < .00001 && value > -.00001){
    return 0;
  } else {
    return value;
  }
}
/** beware of loops!!!
    this function receives signals from other rulers'
    setTangens, so don't call setTangens!!! */
void Ruler::setTangens(double newTangens){
  tangens = smooth(newTangens);
  invTangens = smooth(newTangens ? 1/newTangens : INV_TANGENS_MAX);
  setHorizontal(tangens > -1 && tangens < 1);
  makeRule();
}
/** beware of loops!!!
    this function emits a signal to other rulers'
    setInvTangens */
void Ruler::setInvTangens(const double & newInvTangens){
  invTangens = smooth(newInvTangens);
  tangens = smooth(newInvTangens ? 1/newInvTangens : INV_TANGENS_MAX);
  setHorizontal(tangens > -1 && tangens < 1);
  makeRule();
}
void Ruler::setOrigin(const int & x, const int & y){
  originX = x;
  originY = y;
}
void Ruler::setPoints(const int & x1,const int & y1){
  if( x1!=originX || x2!=originY){
    setTangens((x1!=originX) ? double(y1-originY)/double(x1-originX) : INV_TANGENS_MAX);
  }
}
void Ruler::setPoints(int x1,int y1,int x2,int y2 ){
  if(y1!=y2 || x1!=x2){ // singular, function aborting;
    originX = x2;
    originY = y2;
    setPoints(x1,y1);
  }
}
void Ruler::toggleColor(){
  setPen(QPen(QColor(~ mainLine.pen().color().rgb())));
}




void Ruler::destroy(){
  if(lineList){ destroyLineList(); }
  if(textList){ destroyTextList(); }
  mainLine.setPoints(0,0,0,0);
}
void Ruler::destroyLineList(){            //  this gives a problem when the canvas is destroyed
  if(lineList){                       //  and the lines are already destroyed
    for (int i=0; i<lineListSize; i++){
      delete lineList[i];
    }
    delete[] lineList;
    lineList = 0;
    lineListSize = 0;
  }
}
void Ruler::destroyTextList(){            //  this gives a problem when the canvas is destroyed
  if(textList){                       //  and the lines are already destroyed
    for (int i=0; i<textListSize; i++){
      delete textList[i];
    }
    delete[] textList;
    textList = 0;
    textListSize = 0;
  }
}
void Ruler::makeRule(){
//  cerr << "make rule\n";
  destroy();

  int virHeight = virtualHeight();
  int virWidth  = virtualWidth();
  zero = double(*virtualOriginY)-(*virtualTangens)*double(*virtualOriginX);

  if( ! (*virtualTangens)){
    if( ! inVirtualHeight(*virtualOriginY) )
      return; //       function aborted !!!
    (*vx1)=margin;
    (*vy1)=*virtualOriginY;
    (*vx2)=virWidth;
    (*vy2)=*virtualOriginY;
  } else  {
    int top    = int((*virtualInvTangens)  * (double(margin)-zero));
    int bottom  = int((*virtualInvTangens)  * (double(virHeight)-zero));
    int left    = int((*virtualTangens)    * double(margin)+zero);
    int right  = int((*virtualTangens)    * double(virWidth)+zero);

    int side1, side2;

    if (inVirtualWidth(top)){
      side1=0;
    } else if (inVirtualHeight(left)){
      side1=1;
    } else if (inVirtualHeight(right)){
      side1=2;
    } else {
      side1=100; // nothing found!
    }

    switch(side1) {
      case 0: if (inVirtualHeight(left) && !(left==top && left==margin))
                {side2=1; break;}
      case 1: if (inVirtualHeight(right) && !(right==top && right==margin))
                {side2=2; break;}
      case 2: if (inVirtualWidth(bottom) && !((bottom==left && left==margin) ||
                                    (bottom==right && bottom==virWidth)))
                {side2=3; break;}
      default: // err_outside();
              length = 0;
//              cerr << " virTang " << *virtualTangens << " virInvTang " << *virtualInvTangens << " zero " << zero << endl;
//              cerr << " top " << top << " left " << left << " right " << right << " bottom " << bottom << "\n";
              return;  //       function aborted !!!
    }

    switch (side1){
      case 0:  (*vx1)=top;           (*vy1)=margin;           break;
      case 1:  (*vx1)=margin;        (*vy1)=left;             break;
      case 2:  (*vx1)=virWidth;       (*vy1)=right;           break;
    }
    switch (side2){
      case 1:  (*vx2)=margin;         (*vy2)=left;             break;
      case 2:  (*vx2)=virWidth;        (*vy2)=right;           break;
      case 3:  (*vx2)=bottom;      (*vy2)=virHeight;           break;
    }
//     cerr << " x1 " << x1 << " y1 " << y1 << " x2 " << x2 << " y2 " << y2 << "\n"
//        << " side1 " << side1 << " side2 " << side2 << " \n";
  }
  mainLine.setPoints(x1, height-y1, x2, height-y2);
  length = double(sqrt((double)((x1-x2)*(x1-x2)+(y1-y2)*(y1-y2))));
  makeSpikes();
//  cerr << " voor de tweede: x1 " << x1 << " y1 " << y1 << " x2 " << x2 << " y2 " << y2 << "\n";
  return;
}
void Ruler::makeSpikes(){
  if( ! spikeUnit)
    spikeUnit = 1;
  double x, y, step, xstep, ystep;
  step = double(sqrt((spikeUnit * spikeUnit)/(1+(*virtualTangens)*(*virtualTangens))));
  ystep=step/4;                    // the x-length of the ruler picks
  xstep=(*virtualTangens)*ystep;   // the y-length of the ruler picks

  x = double((*vx1) < (*vx2) ? (*vx1) : (*vx2)); // make sure x is left

  int realStep = int( step * 1000);            // for the next command
  if( ! realStep )
    realStep = 1;

  int distance = int( *virtualOriginX - x) * 1000;
  int center = distance / realStep;
  double onset = double( distance % realStep );

  x += onset / 1000;

  lineListSize = int((length * 1000 - onset) / realStep ) + 1;
  lineList = new RulerLine*[lineListSize];

  textList = new RulerText*[4 + lineListSize]; // 4: begin + center + end + 1st for lineListSize/5
  textListSize = 0;

  double valX, valY;

  for (int i=0; i<lineListSize; i++){
    lineList[i] = new RulerLine(getID(), canvas);
    bool makeText = ( numbers &&
                       ( i == 0 ||
                         i == lineListSize-1 ||
                         ( (center - i) % 5 == 0 &&
                           i > 3 &&
                           i < ( lineListSize - 4 )
                         )
                       )
                     );
    y = (*virtualTangens) * x + zero;
    if(i<(lineListSize-1) || (inVirtualWidth(int(x)) && inVirtualHeight(int(y)))){
      if(horizontal){
        if (makeText) {
          lineList[i]->setPoints(int(x+xstep), height-int(y-ystep), int(x-xstep), height-int(y+ystep));
          textList[textListSize] = new RulerText(getID(), canvas);
          canvas->getAxes()->posToVal((int)x, (int)y, valX, valY);
          textList[textListSize]->setText(QString::number(valX,'g',4));
          textList[textListSize]->setX(x + xstep + 5);
          textList[textListSize]->setY(height-y + ystep - spikeUnit/5 );
          textList[textListSize]->setVisible(visible);
          textListSize++;
        } else {
          lineList[i]->setPoints(int(x), height-int(y), int(x-xstep), height-int(y+ystep));
        }
      } else {
        if (makeText) {
          lineList[i]->setPoints(int(y-ystep), height-int(x+xstep), int(y+ystep), height-int(x-xstep));
          textList[textListSize] = new RulerText(getID(), canvas);
          canvas->getAxes()->posToVal((int)y, height -(int)x, valX, valY);
          textList[textListSize]->setText(QString::number(valY,'g',4));
          double yPos = y - textList[textListSize]->boundingRect().width() - spikeUnit/3 + ystep;
          textList[textListSize]->setX(yPos > 3 ? yPos : 3);
          textList[textListSize]->setY(height - x - textList[textListSize]->boundingRect().height() - xstep);
          textList[textListSize]->setVisible(visible);
          textListSize++;
        } else {
          lineList[i]->setPoints(int(y), height-int(x), int(y+ystep), height-int(x-xstep));
        }
      }
    }
    lineList[i]->setPen(mainLine.pen());
    lineList[i]->setVisible(visible);
    x += step;
  }
}

//-----------------------------------------------------class RulerLine

RulerLine::RulerLine(const int ID, ScientificCanvas * canvas)
  : QCanvasLine(canvas),
    BasicItem(ID)
{}

//-----------------------------------------------------class RulerText

RulerText::RulerText(const int ID, ScientificCanvas * canvas)
  : QCanvasText(canvas),
    BasicItem(ID)
{}

