/***************************************************************************
                          scientificcanvas.cpp  -  description
                             -------------------
    begin                : Wed Jan 15 2003
    copyright            : (C) 2003 by Volker Nannen
    email                : volker@nannen.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "macros.h"

#include "scientificcanvas.h"

#include "project.h"
#include "plot.h"
#include "points.h"

#include <qpen.h>
#include <qpoint.h>
#include <qpointarray.h>


//--------------------------class ScientificCanvasZoom

ScientificCanvasZoom::ScientificCanvasZoom(const int ID, ScientificCanvas * canvas)
  : QCanvasRectangle(canvas),
    BasicItem(ID)
{}

//--------------------------class ScientificCanvas

ScientificCanvas::ScientificCanvas(Project * project,
                         const QString name,
                         QDomElement * element)
  : BasicElement(0, name),
    caption( getElementName() ),
    principal(false),
    mouseLocked(false),
    project(project),
    axes(this),
    plot( new Plot(project, this, name) ),
    zoom(getID(), this)
{
  init();
  xml_read(element);
}
ScientificCanvas::ScientificCanvas(ScientificCanvas * canvas,
                         const QString name)
  : BasicElement(0, name),
    caption( getElementName() ),
    principal(false),
    mouseLocked(false),
    project(canvas->getProject()),
    axes(this),
    plot( new Plot(project, this, name) ),
    zoom(getID(), this)
{
  init();
  xml_read(0);
}
void ScientificCanvas::init(){
  plot->setVScrollBarMode(QScrollView::AlwaysOff);
  plot->setHScrollBarMode(QScrollView::AlwaysOff);
  plot->resizeContents(canvasWidth, canvasHeight);
  plot->setCaption( caption );
  plot->show();
  hideName();
}
ScientificCanvas::~ScientificCanvas(){
  destroy();
  delete plot;
}
void ScientificCanvas::destroy(){
  clear();
}
bool ScientificCanvas::close(){
  if( principal ) project->die();
  else project->deleteGraph(getID());
  return true;
}
bool ScientificCanvas::xml_write(QDomElement * parent){
  canvasWidth = width();
  canvasHeight = height();
  return BasicElement::xml_write( parent );
}
bool ScientificCanvas::xml_read( QDomElement * element ){
  destroy();
  bool result = BasicElement::xml_read(element);
  if(element){
    QDomElement myElement = element->elementsByTagName("Axes").item(0).toElement();
    axes.xml_read( & myElement );
  }
  if(plot) plot->setCaption( caption );
  refresh();
  if(element){
    QDomNodeList nodeList = element->elementsByTagName("Graph");
    for(unsigned int i=0; i<nodeList.count(); i++){
      Points * points = getNewPoints();
      QDomElement myElement = nodeList.item(i).toElement();
      points->xml_read( & myElement );
      points->recalculate();
    }
  }
  setID(getID());
  controlMargin = margin;
  return result;
}
void ScientificCanvas::test(const double number){
  cerr << " canvas " << number << " width: " << canvasWidth
       << " height: " << canvasHeight << endl;
}
void ScientificCanvas::xml_list(){
  BasicElement::xml_list();
  xml_bool("valid", valid, true, HIDE);
  xml_int("width", canvasWidth, 600, EDIT);
  xml_int("height", canvasHeight, 400, EDIT);
  xml_int("margin", margin, MARGIN_DEFAULT, EDIT);
  xml_qstring("bgColor", bgColor, "white", EDIT);
  xml_qstring("caption", caption, caption, HIDE);
  xml_object( & axes, OPEN);
  for(unsigned int i=0; i<screenPoints.size(); i++){
    xml_object( screenPoints[i], HIDE );
  }
}
void ScientificCanvas::setValid(const bool yes){
  valid = yes;
  refresh();
}
void ScientificCanvas::setID(const int newID){
  BasicElement::setID(newID);
  axes.setID(newID);
  zoom.setID(newID);
}
void ScientificCanvas::setSize(const int & width, const int & height){
  canvasWidth = width;
  canvasHeight = height;
}
void ScientificCanvas::refresh(){
//  cerr << getElementName() << "::refresh, h: " << canvasHeight << " w: " << canvasWidth << " \n";
  setBackgroundColor(QColor(bgColor));
  if( canvasWidth != width() || canvasHeight != height() ){
    resize(canvasWidth, canvasHeight);
    axes.setCanvasSize( canvasWidth, canvasHeight );
    if(plot) plot->resizeContents(canvasWidth, canvasHeight);
    valid = false;
  }
  if(margin != controlMargin){
    controlMargin = margin;
    valid = false;
  }
  axes.setMargin(margin);
  axes.refresh();
  if( ! valid ){
    valid = true;
    for(vector<Points *>::iterator iter = screenPoints.begin(); iter != screenPoints.end(); iter++){
      (*iter)->recalculate();
    }
    emit valueChanged();
    if(principal) project->setValid(false);
  }
}
void ScientificCanvas::setFocus(){
  if( plot ) plot->setFocus();
}
void ScientificCanvas::setCaption(const QString newCaption){
  caption = newCaption;
  if( plot ) plot->setCaption(caption);
}
void ScientificCanvas::setAnimated(bool yes){
  if(yes){
  setAdvancePeriod( 50 );
  } else {
  setAdvancePeriod( 0 );
  }
  axes.setAnimated(yes);
}
void ScientificCanvas::setAxesVisible(bool yes){
  axes.setVisible(yes);
}
void ScientificCanvas::toggleColor(const bool yes){
  axes.toggleColor(yes);
}
Points * ScientificCanvas::getNewPoints(const int ID, const QString color){
  Points * newPoints = new Points(0, this);
  newPoints->setID( ID ? ID : getID() );
  newPoints->setContinuous(true);
  newPoints->setVisible(true);
  newPoints->setColor(color);
//  newPoints->setPreserve(true);
  screenPoints.push_back(newPoints);
  return newPoints;
}
void ScientificCanvas::clear(){
  for(vector<Points *>::iterator iter = screenPoints.begin(); iter != screenPoints.end(); iter++){
    delete (*iter);
  }
  screenPoints.clear();
}
void ScientificCanvas::mousePressed(const QMouseEvent * event){
  if(event->button() & QMouseEvent::RightButton){
    zoomX = event->pos().x();
    zoomY = event->pos().y();
    newZoomX = zoomX;
    newZoomY = zoomY;
    if( zoomX < margin ) zoomX = margin;
    if( zoomX > canvasWidth - margin ) zoomX = canvasWidth - margin;
    if( zoomY < margin ) zoomY = margin;
    if( zoomY > canvasHeight - margin ) zoomY = canvasHeight - margin;
    zoom.move( zoomX, zoomY );
    zoom.setSize(0,0);
    zoom.setVisible(true);
    mouseLocked = true;
    update();
  }
}
void ScientificCanvas::mouseMoved(const QMouseEvent * event){
  if( mouseLocked ){
    if( event->state() & QMouseEvent::RightButton ){
      newZoomX = event->pos().x();
      newZoomY = event->pos().y();
      if( newZoomX < margin ) newZoomX = margin;
      if( newZoomX > canvasWidth - margin ) newZoomX = canvasWidth - margin;
      if( newZoomY < margin ) newZoomY = margin;
      if( newZoomY > canvasHeight - margin ) newZoomY = canvasHeight - margin;
      zoom.setSize( event->pos().x() - zoomX, event->pos().y() - zoomY );
      update();
    } else {
      zoom.setVisible(false);
      mouseLocked = false;
      update();
    }
  }
}
void ScientificCanvas::mouseReleased(const QMouseEvent * event){
  if(mouseLocked){
    int newX = event->pos().x();
    int newY = event->pos().y();
    if(    newX != zoomX && newY != zoomY
        && newX < canvasWidth && newX > 0
        && newY < canvasHeight && newY > 0 ){
      double x1, y1, x2, y2, tmp;
      axes.posToVal(zoomX, zoomY, x1, y1);
      axes.posToVal(newZoomX, newZoomY, x2, y2);
      if( x1 > x2 ) {
        tmp = x1; x1=x2; x2=tmp;
      }
      if( y1 < y2 ) {
        tmp = y1; y1=y2; y2=tmp;
      }
      axes.setUpperY(y1);
      axes.setLowerY(y2);
      axes.setLeftX(x1);
      axes.setRightX(x2);
    }
    zoom.setVisible(false);
    mouseLocked = false;
    refresh();
  }
}

