/***************************************************************************
                          timeseries.cpp  -  description
                             -------------------
    begin                : Fri May 10 2002
    copyright            : (C) 2002 by Volker Nannen
    email                : volker@nannen.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "data.h"
#include "timeseries.h"

#include <cstdlib>
#include <cmath>
#include <ctime>

using namespace std;

TimeSeries::TimeSeries(const int ID, const QString name,
                       QDomElement * /*element*/)
  : BasicElement(ID, name)
{
  hideName();
}
TimeSeries::~TimeSeries(){;}
void TimeSeries::xml_list(){
  BasicElement::xml_list();
  xml_int("seed",seed, 0,  HIDE);
  xml_double("scale", scale, 10, EDIT);
}
double TimeSeries::systemNoise(){
  return system_noise * ( ( double(rand()) / RAND_MAX ) -0.5);
}
void TimeSeries::start(const int newSeed){
  seed = newSeed;
  init();
}
//-------------------------------------------------class Autoregression

AutoSeries::AutoSeries(const int ID, QDomElement * element)
  : TimeSeries(ID, "autoregression", element)
{
  xml_read(element);
}
void AutoSeries::xml_list(){
  TimeSeries::xml_list();
  xml_double("offset", offset, 0, EDIT);
  xml_int("steps", steps, 300, EDIT);
  xml_double("system noise", system_noise, 0.5, EDIT);
  xml_double("f", f_param, 0, EDIT);
  xml_double("e",e_param, 0,  EDIT);
  xml_double("d", d_param, 0, EDIT);
  xml_double("c", c_param, 0.5, EDIT);
  xml_double("b", b_param, 0.5, EDIT);
  xml_double("a", a_param, 0, EDIT);
}
void AutoSeries::init(){
  srand( seed ? seed : time(0) );
  y_0 = 0.1 * systemNoise();
  y_1 = 0.1 * systemNoise();
  y_2 = 0.1 * systemNoise();
  y_3 = 0.1 * systemNoise();
  y_4 = 0.1 * systemNoise();
  y_5 = 0.1 * systemNoise();
  myScale = scale / 4;
}
double AutoSeries::next(double /*x*/){
  y_5 = y_4;
  y_4 = y_3;
  y_3 = y_2;
  y_2 = y_1;
  y_1 = y_0;
  y_0 =   a_param
        + b_param * y_1
        + c_param * y_2
        + d_param * y_3
        + e_param * y_4
        + f_param * y_5
        + systemNoise();

  return offset + myScale * y_0;
}

//-------------------------------------------------class Sinus

SinusSeries::SinusSeries(const int ID, QDomElement * element)
  : TimeSeries(ID, "sinus", element)
{
  xml_read(element);
}
void SinusSeries::xml_list(){
  TimeSeries::xml_list();
  xml_int("steps",steps, 3000,  EDIT);
//  xml_double("offset 5", s_5, 0, EDIT);
//  xml_double("frequency 5", f_5, 0, EDIT);
//  xml_double("offset 4", s_4, 0, EDIT);
//  xml_double("frequency 4", f_4, 0, EDIT);
  xml_double("offset 3", s_3, 0, EDIT);
  xml_double("frequency 3", f_3, 0, EDIT);
  xml_double("offset 2", s_2, 0, EDIT);
  xml_double("frequency 2", f_2, 0, EDIT);
  xml_double("offset 1",s_1, 0,  EDIT);
  xml_double("frequency 1", f_1, 0, EDIT);
}
void SinusSeries::init(){
  srand( seed ? seed : time(0) );
}
double SinusSeries::next(double x){
  x *= 6.28318530;
  return scale * (   sin( x * f_1 + s_1)
                     + sin( x * f_2 + s_2)
                     + sin( x * f_3 + s_3)
                     + sin( x * f_4 + s_4)
                     + sin( x * f_5 + s_5)
                    );
}
//-------------------------------------------------class Henon

HenonSeries::HenonSeries(const int ID, QDomElement * element)
  : TimeSeries(ID, "henon system", element)
{
  xml_read(element);
}
void HenonSeries::xml_list(){
  TimeSeries::xml_list();
  xml_double("offset", offset, 0, EDIT);
  xml_int("steps", steps, 50, EDIT);
  xml_double("system noise", system_noise, 0.1, EDIT);
  xml_double("a", a_param, 1.3, EDIT);
  xml_double("b", b_param, 0.2, EDIT);
}
void HenonSeries::init(){
  srand( seed ? seed : time(0) );
  y_0 = 0.1 * systemNoise();
  y_1 = 0.1 * systemNoise();
  z = 0;
  myScale = scale;
}
double HenonSeries::next(double /*x*/){
  y_0 =   1
        - a_param * y_1 * y_1
        + z
        + systemNoise();
  z = b_param * y_1 * y_1;
  y_1 = y_0;

  return offset + myScale * y_0;
}

//-------------------------------------------------class Logistic

LogisticSeries::LogisticSeries(const int ID, QDomElement * element)
  : TimeSeries(ID, "logistic map", element)
{
  xml_read(element);
}
void LogisticSeries::xml_list(){
  TimeSeries::xml_list();
  xml_double("offset", offset, 0, EDIT);
  xml_int("steps", steps, 80, EDIT);
  xml_double("system noise", system_noise, 0.5, EDIT);
  xml_double("a", a_param, 0.5, EDIT);
}
void LogisticSeries::init(){
  srand( seed ? seed : time(0) );
  y_1 = systemNoise();
  myScale = 2.2 * scale;
}
double LogisticSeries::next(double /*x*/){
  y_0 = (a_param * y_1) * ( 1 - y_1 ) + systemNoise();
  y_1 = y_0;

  return offset + myScale * y_0;
}

//-------------------------------------------------class Lorenz

LorenzSeries::LorenzSeries(const int ID, QDomElement * element)
  : TimeSeries(ID, "lorenz system", element)
{
  xml_read(element);
}
void LorenzSeries::xml_list(){
  TimeSeries::xml_list();
  xml_double("offset", offset, 0, EDIT);
  xml_int("steps", steps, 1000, EDIT);
  xml_double("system noise", system_noise, 0.5, EDIT);
  xml_double("c", c_param, 2.667, EDIT);
  xml_double("b", b_param, 28, EDIT);
  xml_double("a", a_param, 10, EDIT);
}
void LorenzSeries::init(){
  srand( seed ? seed : time(0) );
  y_0 = 0.01 * systemNoise();
  z_0 = 0.01 * systemNoise();
  w_0 = 0.01 * systemNoise();
  myScale = scale / 20;
}
double LorenzSeries::next(double /*x*/){
  y_1 = y_0;
  z_1 = z_0;
  w_1 = w_0;
  y_0 = y_1 + 0.01 * a_param * (z_1 - y_1)       + systemNoise();
  z_0 = z_1 + 0.01 * (y_1*(b_param - w_1) - z_1) + systemNoise();
  w_0 = w_0 + 0.01 * (y_1 * z_1 - c_param * w_1) + systemNoise();

  return offset + myScale * y_0;
}

//-------------------------------------------------class Pendulum
PendulumSeries::PendulumSeries(const int ID, QDomElement * element)
  : TimeSeries(ID, "pendulum", element)
{
  xml_read(element);
}
void PendulumSeries::xml_list(){
  TimeSeries::xml_list();
  xml_double("offset", offset, 0, EDIT);
  xml_int("steps", steps, 1000, EDIT);
  xml_double("system noise", system_noise, 0.5, EDIT);
  xml_double("constant", constant, 0.2 , EDIT);
  xml_double("frequency", freq, 0.5, EDIT);
  xml_double("orbit", orbit, 0.67, EDIT);
}
void PendulumSeries::init(){
  srand( seed ? seed : time(0) );
  y_1 = 0.1 * systemNoise();
  z_1 = 0.1 * systemNoise();
  t = 0;
  myScale = scale / 20;
}
double PendulumSeries::next(double /*x*/){
  y_0 = y_1 + 0.1 * z_1;
  z_0 = z_1 + 0.1 * (sin(y_1) - constant * z_1 + freq * cos(orbit * t)) + systemNoise();
  y_1 = y_0;
  z_1 = z_0;
  t += 0.1;

  return offset + myScale * y_0;
}
//-------------------------------------------------class Step
StepSeries::StepSeries(const int ID, QDomElement * element)
  : TimeSeries(ID, "step function", element),
    step(0)
{
  xml_read(element);
}
StepSeries::~StepSeries(){
  if(step) delete[] step;
  step = 0;
}
void StepSeries::xml_list(){
  TimeSeries::xml_list();
  xml_double("offset", offset, 0, EDIT);
  xml_int("steps", steps, 2000, HIDE);
  xml_int("number of steps", stepNumber, 10, EDIT);
}
void StepSeries::init(){
  srand( seed ? seed : time(0) );
  if(stepNumber < 0) stepNumber = 0;
  if(stepNumber > steps/2) stepNumber = steps/2;
  if(step) delete[] step;
  step = new int[stepNumber];
  for(int i=0; i<stepNumber; i++){
    bool search = true;
    while(search){
      step[i] = (int)floor(steps * ( double(rand()) / RAND_MAX ));
      search = false;
      for(int j=0; j<i; j++){
        if (step[j] == step[i]){
          search = true;
          break;
        }
      }
    }
  }
  stepCount = 0;
  a = gaussianNoise() > 0 ? 10 : -10;
  myScale = scale / 10;
}
double StepSeries::next(double /*x*/){
  if( ! step ) return a;
  for(int i=0; i<stepNumber; i++){
    if( step[i] == stepCount ){
      a = -a;
      break;
    }
  }
  stepCount++;
  return offset + myScale * a;
}
//-------------------------------------------------class Thom
ThomSeries::ThomSeries(const int ID, QDomElement * element)
  : TimeSeries(ID, "thom system", element)
{
  xml_read(element);
}
void ThomSeries::xml_list(){
  TimeSeries::xml_list();
  xml_double("offset", offset, 0, EDIT);
  xml_int("steps", steps, 80, EDIT);
  xml_double("system noise", system_noise, 0.5, EDIT);
  xml_double("d", d_param, 0.4, EDIT);
  xml_double("c", c_param, 0.3, EDIT);
  xml_double("b", b_param, 0.3, EDIT);
  xml_double("a", a_param, 0.5, EDIT);
}
void ThomSeries::init(){
  srand( seed ? seed : time(0) );
  y_0 = 0.1 * systemNoise();
  z_0 = 0.1 * systemNoise();
  myScale = scale;
}
double ThomSeries::next(double /*x*/){
  y_0 = a_param * y_1 + b_param * z_1 + systemNoise();
  z_0 = c_param * y_1 + d_param * z_1 + systemNoise();
  y_1 = y_0 - floor(y_0);
  z_1 = z_0 - floor(z_0);

  return offset + myScale * sin(y_0 * 6.283185307);
}
