/***************************************************************************
                          vandermonde.cpp  -  description
                             -------------------
    begin                : ?' ??? 4 2002
    copyright            : (C) 2002 by Volker Nannen
    email                : volker@nannen.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "vandermonde.h"
#include "points.h"
#include "enhancedprogressbar.h"

#include <qdom.h>

#include <m_apm.h>

#include <algorithm>

using namespace std;

VanDerMonde::VanDerMonde(Points * points)
  : BasicElement(0),
    points(points)
{
  setOrder(-1);
  init();
  xml_read(0);
}
VanDerMonde::VanDerMonde(const int ID, Points * points, QDomElement * element)
  : BasicElement(ID),
    points(points)
{
  setOrder(-1);
  init();
  xml_read(element);
}
VanDerMonde::VanDerMonde(const int ID, Points * points, VanDerMonde * other)
  : BasicElement(ID),
    points(points),
    matrix(other->matrix),
    sumX(other->sumX),
    sumY(other->sumY),
    squareY(other->squareY),
    precision(other->precision),
    order(other->order),
    order2(other->order2),
    isProcessed(other->isProcessed),
    isTriangular(other->isTriangular)
{
  init();
}
VanDerMonde::~VanDerMonde(){
  destroy();
}
void VanDerMonde::destroy(){
  if( order ){
    setOrder(-1);
  }
}
void VanDerMonde::init(){
  sumX.setElementName("sumX");
  sumY.setElementName("sumY");
  matrix.setElementName("matrix");
  hideName();
  makeControl();
}
void VanDerMonde::makeFalse(){
  isProcessed = false;
  isTriangular = false;
  makeControl();
}
bool VanDerMonde::xml_read(QDomElement * element){
  bool result = true;
  if(element){
    QDomElement monde = element->elementsByTagName("VanDerMonde").item(0).toElement();
    result = BasicElement::xml_read( & monde);
    QDomNodeList list = monde.elementsByTagName("Matrix");
    bool valid =    isProcessed
                 && makeXML_Element( & sumX, & list)
                 && sumX.getSize() == (order2 + 1)
                 && makeXML_Element( & sumY, & list)
                 && sumY.getSize() == (order + 1)
                 && (
                         ! isTriangular
                      || (
                              makeXML_Element( & matrix, & list)
                           && matrix.getSize() == ( (order + 1) * (order + 2) )
                         )
                    );
    squareY = monde.attribute("squareY", "0");
    if( ! valid ) setOrder(-1);
  }
  makeControl();
  return result;
}
bool VanDerMonde::xml_write(QDomElement * parent){
  bool result = BasicElement::xml_write( parent );
  char out[precision + 100];
  squareY.round(precision);
  squareY.toString(out, -1);
  thisDomElement()->setAttribute("squareY", out);
  return result;
}
void VanDerMonde::makeControl(){
  controlIsTriangular = isTriangular;
  controlIsProcessed = isProcessed;
  controlOrder = order;
}
void VanDerMonde::xml_list(){
  BasicElement::xml_list();
  xml_bool("triangular", isTriangular, false, EDIT);
  xml_bool("process", isProcessed, false, EDIT);
  xml_int("precision", precision, 0, HIDE);
  xml_int("order", order, 0, EDIT);
  xml_int("order 2", order2, 1, HIDE);
  xml_object(&matrix, HIDE);
  xml_object(&sumX, HIDE);
  xml_object(&sumY, HIDE);
}
void VanDerMonde::refresh(){
  if(controlOrder != order){
    setOrder(order);
    emit valueChanged();
  }
  if(controlIsTriangular != isTriangular){
    processData();
    leastSquare(order);
    emit valueChanged();
  } else if (controlIsProcessed != isProcessed){
    processData();
    emit valueChanged();
  }
  makeControl();
}
VanDerMonde & VanDerMonde::operator+=(VanDerMonde & other){
  sumX += other.sumX;
  sumY += other.sumY;
  squareY += other.squareY;
  squareY.round(precision);
  isTriangular = false;
  isProcessed = true;
  makeControl();
  return *this;
}
int VanDerMonde::getSampleSize(){
  return points ? points->size() : 0;
}
bool VanDerMonde::makeParameter(Matrix & parameter, int * precisionArray, int parameterOrder){//  const int newOrder, int newPrecision){
//  cerr << " VDM::makeParameter 1, parameterorder = " << parameterOrder << endl;
  if( ! leastSquare(parameterOrder) ){
    parameter.makeZero();
    return false;
  }
  for(int row = parameterOrder; row >= 0; row--){
    MAPM tmp = MM_Zero;
    for(int col = row+1; col <= parameterOrder; col++){
      tmp += parameter[col] * matrix(row, col);
    }
    if( precisionArray[row] ){
      parameter[row] = (matrix(row, order+1) - tmp).round(precisionArray[row]-1);
    } else {
      parameter[row] = MM_Zero;
    }
  }
  return true;
}
void VanDerMonde::setOrder(int newOrder){
//  cerr << matrix.getSize() << " VDM::setorder, order = " << newOrder << endl;
  if( newOrder < 0 ){
    newOrder = 0;
  } else {
    if(    (newOrder < order)
        || ( ( newOrder == order ) && (controlOrder == order) ) ) return;
  }
  order     = newOrder;
  order2    = order ? order * 2 : 1;
  precision = int(order * 2);
//  cerr << " precsion = 2*order \n";
  if( precision < 10 ) precision = 10;
  sumX.setPrecision(precision);
  sumY.setPrecision(precision);
  matrix.setPrecision(precision);

  sumX.init(order2 + 1);
  sumY.init(order + 1);
  matrix.init(order + 1, order + 2);
  makeFalse();
  squareY = MM_Zero;
}
void VanDerMonde::makeZero(){
  sumX.makeZero();
  sumY.makeZero();
  matrix.makeZero();
  squareY = MM_Zero;                        // for ::variance(VanDerMonde * )
  makeFalse();
}
bool VanDerMonde::leastSquare(const int newOrder, EnhancedProgressBar * dialog){
  if( newOrder > order ) setOrder( newOrder );
  return processData(dialog) && makeTriangular(dialog);
}
bool VanDerMonde::processData(EnhancedProgressBar * dialog){
//  cerr << endl << " VDM::process, order:" << order << " precision: " << precision <<endl;
  if( isProcessed ) return true;
  isTriangular = false;
  makeZero();
  int count = points->size();
  if( ! count ) return false;
  m_apm_cpp_precision(precision);
  MAPM xx, xx_pow, yy;
  EnhancedProgressBar progress(dialog, "Processing sample...", count);
  Matrix powers(order2+1);
  for(int i=0; i<count; i++){
    if ( progress.check(31, i) ) return false;
    powers.makeZero();
    powers[1] = (*points)[i]->getValueX();
    MAPM yy = (*points)[i]->getValueY();
    squareY += yy.ipow(2, precision);
    squareY.round(precision);
    for(int j=2; j <= order2; j <<= 1){
      powers[j] = powers[j>>1].ipow(2, precision);
      for(int k=1; (k<j) && (k+j <= order2); k++){
        powers[j+k] = (powers[j] * powers[k]).round(precision);
      }
    }
    for(int j=1; j<=order2; j++){
      sumX[j] += powers[j];
      sumX[j].round(precision);
    }
    sumY[0] += yy;
    sumY[0].round(precision);
    for(int j=1; j<=order; j++){
      sumY[j] += (powers[j] * yy).round(precision);
      sumY[j].round(precision);
    }
  }
  sumX[0] = (long)count;

  isProcessed = true;
  makeControl();
  return true;
}
bool VanDerMonde::makeTriangular(EnhancedProgressBar * dialog){
  if( isTriangular ) return true;
  for(int i = 0; i <= order; i++){
    for(int j = 0; j <= order; j++){
      matrix(i, j) = sumX[i+j];
    }
    matrix(i, order+1) = sumY[i];
  }
  isTriangular = matrix.makeTriangular(dialog);
  makeControl();

/*  for(int i=0; i<= order; i++){
    char out[100];
    matrix(i, i).toString(out, 3);
    cerr << " [" << i << "]: " << out << " \t";
    if( ! ( (i+1) % 5) ) cerr << endl;
  }*/

  return isTriangular;
}
double VanDerMonde::variance(Matrix & parameter, int parameterOrder){
  if( parameterOrder > parameter.getSize() ) return 0;
  setOrder( parameterOrder );
  if( ! processData() ) return 0;
  MAPM correlation(MM_Zero);
  MAPM squareX(MM_Zero);
  MAPM aa;
  for(int i = 0; i<=parameterOrder; i++){
    correlation += (sumY[i]*parameter[i]).round(precision);
    correlation.round(precision);
  }
  correlation *= 2;
  int parameterOrder2 = parameterOrder * 2;
  for(int i=0; i<=(parameterOrder2); i++){
    aa = MM_Zero;
    int begin = i - parameterOrder;
    if (begin < 0) begin=0;
    for(int j=begin; j < (i-j); j++){
      aa += (parameter[j] * parameter[i-j]).round(precision);
      aa.round(precision);
    }
    aa *= 2;
    if( ! (i & 1) ){
      aa += parameter[i/2].ipow(2, precision);
      aa.round(precision);
    }
    squareX += (sumX[i] * aa).round(precision);
    squareX.round(precision);
  }

//    cerr << " \t squareY: "     << Matrix::toDouble(squareY)
//         << " \t correlation: " << Matrix::toDouble(correlation)
//         << " \t squareX: "     << Matrix::toDouble(squareX) << endl;

  return Matrix::toDouble( ( (squareY - correlation + squareX) / sumX[0])
                                  .round(precision) );
}
